"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Texture, GeomEnums

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.render_target import RenderTarget


class ImageFormatTypes(object):  # pylint: disable=too-few-public-methods
    """ This is a small helper class to prevent pylint errors about the Image
    class not defining the enums. This class just copies the enum properties
    from panda3d's builtin Texture class. """

    T_float = Texture.T_float
    T_unsigned_byte = Texture.T_unsigned_byte
    T_int = Texture.T_int
    T_unsigned_short = Texture.T_unsigned_short
    T_unsigned_int_24_8 = Texture.T_unsigned_int_24_8

    TT_buffer_texture = Texture.TT_buffer_texture
    TT_2d_texture = Texture.TT_2d_texture
    TT_3d_texture = Texture.TT_3d_texture
    TT_cube_map = Texture.TT_cube_map
    TT_cube_map_array = Texture.TT_cube_map_array
    TT_2d_texture_array = Texture.TT_2d_texture_array

    format_format = Texture.format_format
    format_component_type = Texture.format_component_type


class Image(RPObject, Texture, ImageFormatTypes):
    """ This is a wrapper arround the Texture class from Panda3D, which keeps
    track of all images and registers / unregisters them aswell as counting
    the memory used. This is used by all classes instead of pandas builtin
    Texture class. """

    # All registered images
    REGISTERED_IMAGES = []

    # String formats
    FORMAT_MAPPINGS = {
        "R11G11B10": (Texture.T_float, Texture.F_r11_g11_b10),
        "RGBA8": (Texture.T_unsigned_byte, Texture.F_rgba8),
        "RGBA16": (Texture.T_float, Texture.F_rgba16),
        "RGBA32": (Texture.T_float, Texture.F_rgba32),
        "R8": (Texture.T_unsigned_byte, Texture.F_red),
        "R8UI": (Texture.T_unsigned_byte, Texture.F_red),
        "R16": (Texture.T_float, Texture.F_r16),
        "R16UI": (Texture.T_unsigned_short, Texture.F_r16i),
        "R32": (Texture.T_float, Texture.F_r32),
        "R32I": (Texture.T_int, Texture.F_r32i),
    }

    @classmethod
    def create_buffer(cls, name, size, component_format):
        """ Creates a new buffer texture """
        img = cls("ImgBuffer-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_buffer_texture(size, comp_type, comp_format, GeomEnums.UH_static)
        return img

    @classmethod
    def create_counter(cls, name):
        """ Creates a new 1x1 R32I texture to be used as an atomic counter """
        return cls.create_buffer(name, 1, "R32I")

    @classmethod
    def create_2d(cls, name, w, h, component_format):
        """ Creates a new 2D texture """
        img = cls("Img2D-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_2d_texture(w, h, comp_type, comp_format)
        return img

    @classmethod
    def create_2d_array(cls, name, w, h, slices, component_format):
        """ Creates a new 2D-array texture """
        img = cls("Img2DArr-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_2d_texture_array(w, h, slices, comp_type, comp_format)
        return img

    @classmethod
    def create_3d(cls, name, w, h, slices, component_format):
        """ Creates a new 3D texture """
        img = cls("Img3D-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_3d_texture(w, h, slices, comp_type, comp_format)
        return img

    @classmethod
    def create_cube(cls, name, size, component_format):
        """ Creates a new cubemap """
        img = cls("ImgCube-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_cube_map(size, comp_type, comp_format)
        return img

    @classmethod
    def create_cube_array(cls, name, size, num_cubemaps, component_format):
        """ Creates a new cubemap """
        img = cls("ImgCubeArr-" + name)
        comp_type, comp_format = cls.convert_texture_format(component_format)
        img.setup_cube_map_array(size, num_cubemaps, comp_type, comp_format)
        return img

    @classmethod
    def convert_texture_format(cls, comp_type):
        """ Converts a string like 'RGBA8' to a texture type and format """
        return cls.FORMAT_MAPPINGS[comp_type]

    def __init__(self, name):
        """ Internal method to create a new image """
        RPObject.__init__(self, name)
        Texture.__init__(self, name)
        Image.REGISTERED_IMAGES.append(self)
        self.set_clear_color(0)
        self.clear_image()
        self.sort = RenderTarget.CURRENT_SORT

    def __del__(self):
        """ Destroys the image """
        self.warn("Image destructor not implemented yet")

    def write(self, pth):
        """ Writes the image to disk """
        Globals.base.graphicsEngine.extract_texture_data(self, Globals.base.win.gsg)
        if self.get_texture_type() in [Texture.TT_3d_texture, Texture.TT_cube_map]:
            Texture.write(self, "#_" + pth, 0, 0, True, False)
        else:
            Texture.write(self, pth)



import sys
import cv2

from cuda import cudart

import numpy as np
import cupy as cp

import pyrr
import glfw

from OpenGL.GL import *  # noqa F403
import OpenGL.GL.shaders

# require:
# 1. pip install cupy-cuda12x
# 2. CUDA-Python
# 3. PyOpenGL
# 4. pyrr
# 5. glfw


def format_cudart_err(err):
    return (
        f"{cudart.cudaGetErrorName(err)[1].decode('utf-8')}({int(err)}): "
        f"{cudart.cudaGetErrorString(err)[1].decode('utf-8')}"
    )


def check_cudart_err(args):
    if isinstance(args, tuple):
        assert len(args) >= 1
        err = args[0]
        if len(args) == 1:
            ret = None
        elif len(args) == 2:
            ret = args[1]
        else:
            ret = args[1:]
    else:
        err = args
        ret = None

    assert isinstance(err, cudart.cudaError_t), type(err)
    if err != cudart.cudaError_t.cudaSuccess:
        raise RuntimeError(format_cudart_err(err))

    return ret


class CudaOpenGLMappedBuffer:
    def __init__(self, gl_buffer, flags=0):
        self._gl_buffer = int(gl_buffer)
        self._flags = int(flags)

        self._graphics_ressource = None
        self._cuda_buffer = None

        self.register()

    @property
    def gl_buffer(self):
        return self._gl_buffer

    @property
    def cuda_buffer(self):
        assert self.mapped
        return self._cuda_buffer

    @property
    def graphics_ressource(self):
        assert self.registered
        return self._graphics_ressource

    @property
    def registered(self):
        return self._graphics_ressource is not None

    @property
    def mapped(self):
        return self._cuda_buffer is not None

    def __enter__(self):
        return self.map()

    def __exit__(self, exc_type, exc_value, trace):
        self.unmap()
        return False

    def __del__(self):
        self.unregister()

    def register(self):
        if self.registered:
            return self._graphics_ressource
        self._graphics_ressource = check_cudart_err(cudart.cudaGraphicsGLRegisterBuffer(self._gl_buffer, self._flags))
        return self._graphics_ressource

    def unregister(self):
        if not self.registered:
            return self
        self.unmap()
        self._graphics_ressource = check_cudart_err(cudart.cudaGraphicsUnregisterResource(self._graphics_ressource))
        return self

    def map(self, stream=None):
        if not self.registered:
            raise RuntimeError("Cannot map an unregistered buffer.")
        if self.mapped:
            return self._cuda_buffer

        check_cudart_err(cudart.cudaGraphicsMapResources(1, self._graphics_ressource, stream))

        ptr, size = check_cudart_err(cudart.cudaGraphicsResourceGetMappedPointer(self._graphics_ressource))

        self._cuda_buffer = cp.cuda.MemoryPointer(cp.cuda.UnownedMemory(ptr, size, self), 0)

        return self._cuda_buffer

    def unmap(self, stream=None):
        if not self.registered:
            raise RuntimeError("Cannot unmap an unregistered buffer.")
        if not self.mapped:
            return self

        self._cuda_buffer = check_cudart_err(cudart.cudaGraphicsUnmapResources(1, self._graphics_ressource, stream))

        return self


class CudaOpenGLMappedArray(CudaOpenGLMappedBuffer):
    def __init__(self, dtype, shape, gl_buffer, flags=0, strides=None, order='C'):
        super().__init__(gl_buffer, flags)
        self._dtype = dtype
        self._shape = shape
        self._strides = strides
        self._order = order

    @property
    def cuda_array(self):
        assert self.mapped
        return cp.ndarray(
            shape=self._shape,
            dtype=self._dtype,
            strides=self._strides,
            order=self._order,
            memptr=self._cuda_buffer,
        )

    def map(self, *args, **kwargs):
        super().map(*args, **kwargs)
        return self.cuda_array


VERTEX_SHADER = """
#version 330
in vec3 position;
uniform mat4 transform;
void main() {
    gl_Position = transform * vec4(position, 1.0f);
}
"""

FRAGMENT_SHADER = """
#version 330
out vec4 outColor;
void main() {
    outColor = vec4(1.0f, 0.0f, 0.0f, 1.0f);
}
"""


def setup_buffers(Nx, Ny):
    max_vertices = Nx * Ny
    max_triangles = 2 * (Nx - 1) * (Ny - 1)

    ftype = np.float32
    itype = np.uint32

    vertex_bytes = 3 * max_vertices * ftype().nbytes
    index_bytes = 3 * max_triangles * itype().nbytes

    flags = cudart.cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsWriteDiscard

    VBO = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, VBO)
    glBufferData(GL_ARRAY_BUFFER, vertex_bytes, None, GL_DYNAMIC_DRAW)
    vertex_buffer = CudaOpenGLMappedArray(ftype, (Ny, Nx, 3), VBO, flags)

    EBO = glGenBuffers(1)
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, EBO)
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, index_bytes, None, GL_DYNAMIC_DRAW)
    index_buffer = CudaOpenGLMappedArray(itype, (Ny - 1, Nx - 1, 2, 3), EBO, flags)

    Ix = cp.arange(Nx - 1)[None, :]
    Iy = cp.arange(Ny - 1)[:, None]

    I = index_buffer.map()

    I[..., 0, 0] = Iy * Nx + Ix
    I[..., 0, 1] = Iy * Nx + Ix + Nx
    I[..., 0, 2] = Iy * Nx + Ix + Nx + 1
    I[..., 1, 0] = Iy * Nx + Ix + Nx + 1
    I[..., 1, 1] = Iy * Nx + Ix + 1
    I[..., 1, 2] = Iy * Nx + Ix

    index_buffer.unmap()

    return vertex_buffer, index_buffer, max_triangles


def update_vertices(t, vertex_buffer):
    r = (t / 10.0) % 1.0
    r = (1 - r) if r > 0.5 else r

    with vertex_buffer as V:
        (Ny, Nx, _) = V.shape
        # np_image = cp.asnumpy(V)
        # cv2.imshow("window", np_image)
        # cv2.waitKey(1)
        θ = cp.linspace(0, np.pi, Nx)[None, :]
        φ = cp.linspace(0, 2 * np.pi, Ny)[:, None]
        V[..., 0] = r * cp.cos(φ) * cp.sin(θ)
        V[..., 1] = r * cp.sin(φ) * cp.sin(θ)
        V[..., 2] = r * cp.cos(θ)


def main():
    if not glfw.init():
        return
    title = "CuPy Cuda/OpenGL interop example"
    window = glfw.create_window(800, 800, title, None, None)

    if not window:
        glfw.terminate()
        return

    glfw.make_context_current(window)
    glfw.swap_interval(0)

    shader = OpenGL.GL.shaders.compileProgram(
        OpenGL.GL.shaders.compileShader(VERTEX_SHADER, GL_VERTEX_SHADER),
        OpenGL.GL.shaders.compileShader(FRAGMENT_SHADER, GL_FRAGMENT_SHADER),
    )
    positionLoc = glGetAttribLocation(shader, "position")
    transformLoc = glGetUniformLocation(shader, "transform")

    Nx, Ny = 32, 32
    vertex_buffer, index_buffer, Ntriangles = setup_buffers(Nx, Ny)

    fps = 0
    nframes = 0
    last_time = glfw.get_time()

    glUseProgram(shader)
    glEnable(GL_DEPTH_TEST)

    while not glfw.window_should_close(window):
        t = glfw.get_time()
        dt = t - last_time
        if dt >= 1.0:
            fps = nframes / dt
            last_time = t
            nframes = 0

        update_vertices(t, vertex_buffer)

        width, height = glfw.get_window_size(window)
        glViewport(0, 0, width, height)

        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

        rot_x = pyrr.Matrix44.from_x_rotation(0)
        rot_y = pyrr.Matrix44.from_y_rotation(t)

        glUniformMatrix4fv(transformLoc, 1, GL_FALSE, rot_x * rot_y)

        glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer.gl_buffer)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index_buffer.gl_buffer)
        glEnableVertexAttribArray(positionLoc)
        glVertexAttribPointer(positionLoc, 3, GL_FLOAT, GL_FALSE, 12, ctypes.c_void_p(0))
        glDrawElements(GL_TRIANGLES, int(3 * Ntriangles), GL_UNSIGNED_INT, None)

        glfw.swap_buffers(window)
        glfw.poll_events()
        glfw.set_window_title(window, f"{title} ({fps:.1f} fps)")
        nframes += 1

    index_buffer.unregister()
    vertex_buffer.unregister()

    glfw.terminate()


if __name__ == "__main__":
    sys.exit(main())



import logging
import sys
import time
from typing import Optional, Union, Tuple

import gltf
from direct.gui.OnscreenImage import OnscreenImage
from direct.showbase import ShowBase
from panda3d.bullet import BulletDebugNode
from panda3d.core import AntialiasAttrib, loadPrcFileData, LineSegs, PythonCallbackObject, Vec3, NodePath

from metadrive.constants import RENDER_MODE_OFFSCREEN, RENDER_MODE_NONE, RENDER_MODE_ONSCREEN, EDITION, CamMask, \
    BKG_COLOR
from metadrive.engine.asset_loader import initialize_asset_loader, close_asset_loader, randomize_cover, get_logo_file
from metadrive.engine.core.collision_callback import collision_callback
from metadrive.engine.core.force_fps import ForceFPS
from metadrive.engine.core.image_buffer import ImageBuffer
from metadrive.engine.core.light import Light
from metadrive.engine.core.onscreen_message import ScreenMessage
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.engine.core.sky_box import SkyBox
from metadrive.engine.core.terrain import Terrain
from metadrive.utils.utils import is_mac, setup_logger


def _suppress_warning():
    loadPrcFileData("", "notify-level-glgsg fatal")
    loadPrcFileData("", "notify-level-pgraph fatal")
    loadPrcFileData("", "notify-level-pnmimage fatal")
    loadPrcFileData("", "notify-level-thread fatal")
    loadPrcFileData("", "notify-level-bullet fatal")
    loadPrcFileData("", "notify-level-display fatal")


def _free_warning():
    loadPrcFileData("", "notify-level-glgsg debug")
    # loadPrcFileData("", "notify-level-pgraph debug")  # press 4 to use toggle analyze to do this
    loadPrcFileData("", "notify-level-display debug")  # press 4 to use toggle analyze to do this
    loadPrcFileData("", "notify-level-pnmimage debug")
    loadPrcFileData("", "notify-level-thread debug")


def attach_cover_image(window_width, window_height):
    cover_file_path = randomize_cover()
    image = OnscreenImage(image=cover_file_path, pos=(0, 0, 0), scale=(1, 1, 1))
    if window_width > window_height:
        scale = window_width / window_height
    else:
        scale = window_height / window_width
    image.set_scale((scale, 1, scale))
    image.setTransparency(True)
    return image


def attach_logo(engine):
    cover_file_path = get_logo_file()
    image = OnscreenImage(image=cover_file_path)
    scale = 0.075
    image.set_scale((scale * 3, 1, scale))
    image.set_pos((0.8325 * engine.w_scale, 0, -0.94 * engine.h_scale))
    image.set_antialias(AntialiasAttrib.MMultisample)
    image.setTransparency(True)
    return image


class EngineCore(ShowBase.ShowBase):
    DEBUG = False
    global_config = None  # global config can exist before engine initialization
    loadPrcFileData("", "window-title {}".format(EDITION))
    loadPrcFileData("", "framebuffer-multisample 1")
    loadPrcFileData("", "multisamples 8")
    loadPrcFileData("", "bullet-filter-algorithm groups-mask")
    loadPrcFileData("", "audio-library-name null")
    loadPrcFileData("", "model-cache-compressed-textures 1")
    loadPrcFileData("", "textures-power-2 none")

    # loadPrcFileData("", "transform-cache 0")
    # loadPrcFileData("", "state-cache 0")
    loadPrcFileData("", "garbage-collect-states 0")

    # loadPrcFileData("", "allow-incomplete-render #t")
    # loadPrcFileData("", "# even-animation #t")

    # loadPrcFileData("", " framebuffer-srgb truein")
    # loadPrcFileData("", "geom-cache-size 50000")

    # v-sync, it seems useless
    # loadPrcFileData("", "sync-video 1")

    # for debug use
    # loadPrcFileData("", "gl-version 3 2")

    def __init__(self, global_config):
        # if EngineCore.global_config is not None:
        #     # assert global_config is EngineCore.global_config, \
        #     #     "No allowed to change ptr of global config, which may cause issue"
        #     pass
        # else:
        EngineCore.global_config = global_config
        self.mode = global_config["_render_mode"]
        if self.global_config["pstats"]:
            # pstats debug provided by panda3d
            loadPrcFileData("", "want-pstats 1")

        # Setup onscreen render
        if self.global_config["use_render"]:
            assert self.mode == RENDER_MODE_ONSCREEN, "Render mode error"
            # Warning it may cause memory leak, Pand3d Official has fixed this in their master branch.
            # You can enable it if your panda version is latest.
            if self.global_config["multi_thread_render"] and not self.use_render_pipeline:
                # multi-thread render, accelerate simulation
                loadPrcFileData("", "threading-model {}".format(self.global_config["multi_thread_render_mode"]))
        else:
            self.global_config["show_coordinates"] = False
            if self.global_config["image_observation"]:
                assert self.mode == RENDER_MODE_OFFSCREEN, "Render mode error"
                if self.global_config["multi_thread_render"] and not self.use_render_pipeline:
                    # render-pipeline can not work with multi-thread rendering
                    loadPrcFileData("", "threading-model {}".format(self.global_config["multi_thread_render_mode"]))
            else:
                assert self.mode == RENDER_MODE_NONE, "Render mode error"
                if self.global_config["show_interface"]:
                    # Disable useless camera capturing in none mode
                    self.global_config["show_interface"] = False

        if is_mac() and (self.mode == RENDER_MODE_OFFSCREEN):  # Mac don't support offscreen rendering
            self.mode = RENDER_MODE_ONSCREEN

        loadPrcFileData("", "win-size {} {}".format(*self.global_config["window_size"]))

        if self.use_render_pipeline:
            from metadrive.render_pipeline.rpcore import RenderPipeline
            from metadrive.render_pipeline.rpcore.rpobject import RPObject
            if not self.global_config["debug"]:
                RPObject.set_output_level("warning")  # warning
            self.render_pipeline = RenderPipeline()
            self.render_pipeline.pre_showbase_init()
            # disable it, as some model errors happen!
            loadPrcFileData("", "gl-immutable-texture-storage 0")
        else:
            self.render_pipeline = None

        # Setup some debug options
        # if self.global_config["headless_machine_render"]:
        #     # headless machine support
        #     # loadPrcFileData("", "load-display  pandagles2")
        #     # no further actions will be applied now!
        #     pass
        if self.global_config["debug"]:
            # debug setting
            EngineCore.DEBUG = True
            if self.global_config["debug_panda3d"]:
                _free_warning()
            setup_logger(debug=True)
            self.accept("1", self.toggleDebug)
            self.accept("2", self.toggleWireframe)
            self.accept("3", self.toggleTexture)
            self.accept("4", self.toggleAnalyze)
            self.accept("5", self.reload_shader)
        else:
            # only report fatal error when debug is False
            _suppress_warning()
            # a special debug mode
            if self.global_config["debug_physics_world"]:
                self.accept("1", self.toggleDebug)
                self.accept("4", self.toggleAnalyze)

        if not self.global_config["disable_model_compression"] and not self.use_render_pipeline:
            loadPrcFileData("", "compressed-textures 1")  # Default to compress

        super(EngineCore, self).__init__(windowType=self.mode)
        if self.use_render_pipeline and self.mode != RENDER_MODE_NONE:
            self.render_pipeline.create(self)

        # main_window_position = (0, 0)
        if self.mode == RENDER_MODE_ONSCREEN:
            h = self.pipe.getDisplayHeight()
            w = self.pipe.getDisplayWidth()
            if self.global_config["window_size"][0] > 0.9 * w or self.global_config["window_size"][1] > 0.9 * h:
                old_scale = self.global_config["window_size"][0] / self.global_config["window_size"][1]
                new_w = int(min(0.9 * w, 0.9 * h * old_scale))
                new_h = int(min(0.9 * h, 0.9 * w / old_scale))
                self.global_config["window_size"] = tuple([new_w, new_h])
                from panda3d.core import WindowProperties
                props = WindowProperties()
                props.setSize(self.global_config["window_size"][0], self.global_config["window_size"][1])
                self.win.requestProperties(props)
                logging.warning(
                    "Since your screen is too small ({}, {}), we resize the window to {}.".format(
                        w, h, self.global_config["window_size"]
                    )
                )
            # main_window_position = (
            #     (w - self.global_config["window_size"][0]) / 2, (h - self.global_config["window_size"][1]) / 2
            # )

        # screen scale factor
        self.w_scale = max(self.global_config["window_size"][0] / self.global_config["window_size"][1], 1)
        self.h_scale = max(self.global_config["window_size"][1] / self.global_config["window_size"][0], 1)

        if self.mode == RENDER_MODE_ONSCREEN:
            self.disableMouse()

        if not self.global_config["debug_physics_world"] \
                and (self.mode in [RENDER_MODE_ONSCREEN, RENDER_MODE_OFFSCREEN]):
            initialize_asset_loader(self)
            gltf.patch_loader(self.loader)
            if not self.use_render_pipeline:
                # Display logo
                if self.mode == RENDER_MODE_ONSCREEN and (not self.global_config["debug"]):
                    if self.global_config["show_logo"]:
                        self._window_logo = attach_logo(self)
                    self._loading_logo = attach_cover_image(
                        window_width=self.get_size()[0], window_height=self.get_size()[1]
                    )
                    for i in range(5):
                        self.graphicsEngine.renderFrame()
                    self.taskMgr.add(self.remove_logo, "remove _loading_logo in first frame")

        self.closed = False

        # add element to render and pbr render, if is exists all the time.
        # these element will not be removed when clear_world() is called
        self.pbr_render = self.render.attachNewNode("pbrNP")

        # attach node to this root whose children nodes will be clear after calling clear_world()
        self.worldNP = self.render.attachNewNode("world_np")
        self.origin = self.worldNP

        # same as worldNP, but this node is only used for render gltf model with pbr material
        self.pbr_worldNP = self.pbr_render.attachNewNode("pbrNP")
        self.debug_node = None

        # some render attribute
        self.pbrpipe = None
        self.world_light = None
        self.common_filter = None

        # physics world
        self.physics_world = PhysicsWorld(self.global_config["debug_static_world"])

        # collision callback
        self.physics_world.dynamic_world.setContactAddedCallback(PythonCallbackObject(collision_callback))

        # for real time simulation
        self.force_fps = ForceFPS(self)

        # init terrain
        self.terrain = Terrain(self.global_config["show_terrain"], self)
        # self.terrain.attach_to_world(self.render, self.physics_world)

        self.sky_box = None

        # init other world elements
        if self.mode != RENDER_MODE_NONE:
            if self.use_render_pipeline:
                self.pbrpipe = None
                if self.global_config["daytime"] is not None:
                    self.render_pipeline.daytime_mgr.time = self.global_config["daytime"]
            else:
                from metadrive.engine.core.our_pbr import OurPipeline
                self.pbrpipe = OurPipeline(
                    render_node=None,
                    window=None,
                    camera_node=None,
                    msaa_samples=16,
                    max_lights=8,
                    use_normal_maps=False,
                    use_emission_maps=True,
                    exposure=1.0,
                    enable_shadows=False,
                    enable_fog=False,
                    use_occlusion_maps=False
                )
                self.pbrpipe.render_node = self.pbr_render
                self.pbrpipe.render_node.set_antialias(AntialiasAttrib.M_auto)
                self.pbrpipe._recompile_pbr()
                # self.pbrpipe.manager.cleanup()
                #
                # # filter
                # from direct.filter.CommonFilters import CommonFilters
                # self.common_filter = CommonFilters(self.win, self.cam)
                # self.common_filter.set_gamma_adjust(0.8)

                self.sky_box = SkyBox(not self.global_config["show_skybox"])
                self.sky_box.attach_to_world(self.render, self.physics_world)

                self.world_light = Light(self.global_config)
                self.world_light.attach_to_world(self.render, self.physics_world)
                self.render.setLight(self.world_light.direction_np)
                self.render.setLight(self.world_light.ambient_np)
                self.render.setShaderAuto()
                self.render.setAntialias(AntialiasAttrib.MAuto)

            # set main cam
            self.cam.node().setCameraMask(CamMask.MainCam)
            self.cam.node().getDisplayRegion(0).setClearColorActive(True)
            self.cam.node().getDisplayRegion(0).setClearColor(BKG_COLOR)
            lens = self.cam.node().getLens()

            lens.setFov(self.global_config["camera_fov"])

            # ui and render property
            if self.global_config["show_fps"]:
                self.setFrameRateMeter(True)

            # onscreen message
            self.on_screen_message = ScreenMessage(debug=self.DEBUG) if self.mode == RENDER_MODE_ONSCREEN else None
            self._show_help_message = False
            self._episode_start_time = time.time()

            self.accept("h", self.toggle_help_message)
            self.accept("f", self.force_fps.toggle)
            self.accept("escape", sys.exit)
        else:
            self.on_screen_message = None

        # task manager
        self.taskMgr.remove("audioLoop")

        self.coordinate_line = []
        if self.global_config["show_coordinates"]:
            self.show_coordinates()

        # create sensors
        self.sensors = {}
        self.setup_sensors()

    def render_frame(self, text: Optional[Union[dict, str]] = None):
        """
        The real rendering is conducted by the igLoop task maintained by panda3d.
        Frame will be drawn and refresh, when taskMgr.step() is called.
        This function is only used to pass the message that needed to be printed in the screen to underlying renderer.
        :param text: A dict containing key and values or a string.
        :return: None
        """

        if self.on_screen_message is not None:
            self.on_screen_message.render(text)
        if self.mode != RENDER_MODE_NONE and self.sky_box is not None:
            self.sky_box.step()

    def step_physics_world(self):
        dt = self.global_config["physics_world_step_size"]
        self.physics_world.dynamic_world.doPhysics(dt, 1, dt)

    def _debug_mode(self):
        debugNode = BulletDebugNode("Debug")
        debugNode.showWireframe(True)
        debugNode.showConstraints(True)
        debugNode.showBoundingBoxes(False)
        debugNode.showNormals(True)
        debugNP = self.render.attachNewNode(debugNode)
        self.physics_world.dynamic_world.setDebugNode(debugNP.node())
        self.debug_node = debugNP
        # TODO: debugNP is not registered.

    def toggleAnalyze(self):
        self.worldNP.analyze()
        print(self.physics_world.report_bodies())
        # self.worldNP.ls()

    def toggleDebug(self):
        if self.debug_node is None:
            self._debug_mode()
        if self.debug_node.isHidden():
            self.debug_node.show()
        else:
            self.debug_node.hide()

    def report_body_nums(self, task):
        logging.debug(self.physics_world.report_bodies())
        return task.done

    def close_world(self):
        self.taskMgr.stop()
        # It will report a warning said AsynTaskChain is created when taskMgr.destroy() is called but a new showbase is
        # created.
        logging.debug(
            "Before del taskMgr: task_chain_num={}, all_tasks={}".format(
                self.taskMgr.mgr.getNumTaskChains(), self.taskMgr.getAllTasks()
            )
        )
        self.taskMgr.destroy()
        logging.debug(
            "After del taskMgr: task_chain_num={}, all_tasks={}".format(
                self.taskMgr.mgr.getNumTaskChains(), self.taskMgr.getAllTasks()
            )
        )
        self.physics_world.dynamic_world.clearContactAddedCallback()
        self.physics_world.destroy()
        self.destroy()
        close_asset_loader()
        EngineCore.global_config = None

        import sys
        if sys.version_info >= (3, 0):
            import builtins
        else:
            import __builtin__ as builtins
        if hasattr(builtins, "base"):
            del builtins.base

    def clear_world(self):
        self.worldNP.removeNode()
        self.pbr_worldNP.removeNode()

    def toggle_help_message(self):
        if self.on_screen_message:
            self.on_screen_message.toggle_help_message()

    def draw_line_2d(self, start_p, end_p, color, thickness: float):
        """
        Draw line use LineSegs coordinates system. Since a resolution problem is solved, the point on screen should be
        described by [horizontal ratio, vertical ratio], each of them are ranged in [-1, 1]
        :param start_p: 2d vec
        :param end_p: 2d vec
        :param color: 4d vec, line color
        :param thickness: line thickness
        """
        line_seg = LineSegs("interface")
        line_seg.setColor(*color)
        line_seg.moveTo(start_p[0] * self.w_scale, 0, start_p[1] * self.h_scale)
        line_seg.drawTo(end_p[0] * self.w_scale, 0, end_p[1] * self.h_scale)
        line_seg.setThickness(thickness)
        line_np = self.aspect2d.attachNewNode(line_seg.create(False))
        # TODO: line_np is not registered.
        return line_np

    def remove_logo(self, task):
        alpha = self._loading_logo.getColor()[-1]
        if alpha < 0.1:
            self._loading_logo.destroy()
            return task.done
        else:
            new_alpha = alpha - 0.08
            self._loading_logo.setColor((1, 1, 1, new_alpha))
            return task.cont

    def draw_line_3d(self, start_p: Union[Vec3, Tuple], end_p: Union[Vec3, Tuple], color, thickness: float):
        assert self.mode == RENDER_MODE_ONSCREEN, "Can not call this API in render mode: {}".format(self.mode)
        start_p = [*start_p]
        end_p = [*end_p]
        start_p[1] *= 1
        end_p[1] *= 1
        line_seg = LineSegs("interface")
        line_seg.setColor(*color)
        line_seg.moveTo(Vec3(*start_p))
        line_seg.drawTo(Vec3(*end_p))
        line_seg.setThickness(thickness)
        np = NodePath(line_seg.create(False))
        # np.reparentTo(self.render)
        return np

    def show_coordinates(self):
        if len(self.coordinate_line) > 0:
            return
        # x direction = red
        np_x = self.draw_line_3d(Vec3(0, 0, 0.1), Vec3(100, 0, 0.1), color=[1, 0, 0, 1], thickness=2)
        np_x.reparentTo(self.render)
        # y direction = blue
        np_y = self.draw_line_3d(Vec3(0, 0, 0.1), Vec3(0, 50, 0.1), color=[0, 1, 0, 1], thickness=2)
        np_y.reparentTo(self.render)
        self.coordinate_line.append(np_x)
        self.coordinate_line.append(np_y)

    def remove_coordinates(self):
        for line in self.coordinate_line:
            line.detachNode()
            line.removeNode()

    def set_coordinates_indicator_pos(self, pos):
        if len(self.coordinate_line) == 0:
            return
        for line in self.coordinate_line:
            line.setPos(pos[0], pos[1], 0)

    @property
    def use_render_pipeline(self):
        return self.global_config["render_pipeline"] and not self.mode == RENDER_MODE_NONE

    def reload_shader(self):
        if self.render_pipeline is not None:
            self.render_pipeline.reload_shaders()

    def setup_sensors(self):
        for sensor_id, sensor_cfg in self.global_config["sensors"].items():
            if sensor_id == "main_camera":
                # It is added when initializing main_camera
                continue
            cls = sensor_cfg[0]
            args = sensor_cfg[1:]
            if issubclass(cls, ImageBuffer):
                self.add_image_sensor(sensor_id, cls, args)
            else:
                raise ValueError("TODO: Add other sensor here")

    def get_sensor(self, sensor_id):
        if sensor_id not in self.sensors:
            raise ValueError("Can not get {}, available sensors: {}".format(sensor_id, self.sensors.keys()))
        return self.sensors[sensor_id]

    def add_image_sensor(self, name: str, cls, args):
        if self.global_config["image_on_cuda"] and name == self.global_config["vehicle_config"]["image_source"]:
            sensor = cls(self, *args, cuda=True)
        else:
            sensor = cls(self, *args, cuda=False)
        assert isinstance(sensor, ImageBuffer), "This API is for adding image sensor"
        self.sensors[name] = sensor



"""
Converts the cubemap to a spherical one
"""

from __future__ import division, print_function

from panda3d.core import *  # noqa
load_prc_file_data("", "textures-power-2 none")

import direct.directbase.DirectStart  # noqa

cubemap = loader.load_cube_map("../../../../data/default_cubemap/source/#.jpg")
w, h = 4096, 2048

cshader = Shader.make_compute(
    Shader.SL_GLSL, """
#version 430
layout (local_size_x = 16, local_size_y = 16, local_size_z = 1) in;

uniform samplerCube SourceTex;
uniform writeonly image2D DestTex;
#define M_PI 3.1415926535897932384626433
#define TWO_PI 6.2831853071795864769252867

// Converts a normalized spherical coordinate (r = 1) to cartesian coordinates
vec3 spherical_to_vector(float theta, float phi) {
    float sin_theta = sin(theta);
    return normalize(vec3(
        sin_theta * cos(phi),
        sin_theta * sin(phi),
        cos(theta)
    ));
}

// Fixes the cubemap direction
vec3 fix_cubemap_coord(vec3 coord) {
    return normalize(coord.xzy * vec3(1,-1,1));
}

void main() {
    ivec2 dimensions = imageSize(DestTex).xy;
    ivec2 coord = ivec2(gl_GlobalInvocationID.xy);
    float theta = (coord.x + 0.5) / float(dimensions.x) * TWO_PI;
    float phi = (dimensions.y - coord.y - 0.5) / float(dimensions.y) * M_PI;
    vec3 v = spherical_to_vector(phi, theta);
    v = fix_cubemap_coord(v);
    vec4 color = texture(SourceTex, v);
    imageStore(DestTex, coord, vec4(color));
}
"""
)

dest_tex = Texture("")
dest_tex.setup_2d_texture(w, h, Texture.T_float, Texture.F_rgba16)

print("Converting to spherical coordinates ..")
np = NodePath("np")
np.set_shader(cshader)
np.set_shader_input("SourceTex", cubemap)
np.set_shader_input("DestTex", dest_tex)
attr = np.get_attrib(ShaderAttrib)
base.graphicsEngine.dispatch_compute((w // 16, h // 16, 1), attr, base.win.gsg)

print("Extracting data ..")
base.graphicsEngine.extract_texture_data(dest_tex, base.win.gsg)

print("Writing texture ..")
dest_tex.write("envmap.png")



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import SamplerState

# Load the plugin api
from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin
from metadrive.render_pipeline.rpcore.loader import RPLoader

from .color_correction_stage import ColorCorrectionStage
from .auto_exposure_stage import AutoExposureStage
from .manual_exposure_stage import ManualExposureStage
from .sharpen_stage import SharpenStage
from .tonemapping_stage import TonemappingStage


class Plugin(BasePlugin):

    name = "Color Correction"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = (
        "This plugin adds support for color correction, vignetting, "
        "chromatic abberation and tonemapping. It also controls the "
        "camera parameters."
    )
    version = "1.4"

    def on_stage_setup(self):
        self.stage = self.create_stage(ColorCorrectionStage)
        self.tonemapping_stage = self.create_stage(TonemappingStage)

        if self.get_setting("use_sharpen"):
            self.sharpen_stage = self.create_stage(SharpenStage)
            self.sharpen_stage.sharpen_twice = self.get_setting("sharpen_twice")

        if not self.get_setting("manual_camera_parameters"):
            self.exposure_stage = self.create_stage(AutoExposureStage)
        else:
            self.exposure_stage = self.create_stage(ManualExposureStage)  # noqa # pylint: disable=redefined-variable-type

    def on_pipeline_created(self):
        self.load_lut()
        self.load_grain()

    def load_lut(self):
        """ Loads the color correction lookup table (LUT) """
        lut_path = self.get_resource(self.get_setting("color_lut"))
        lut = RPLoader.load_sliced_3d_texture(lut_path, 64)
        lut.set_wrap_u(SamplerState.WM_clamp)
        lut.set_wrap_v(SamplerState.WM_clamp)
        lut.set_wrap_w(SamplerState.WM_clamp)
        lut.set_minfilter(SamplerState.FT_linear)
        lut.set_magfilter(SamplerState.FT_linear)
        lut.set_anisotropic_degree(0)
        self.tonemapping_stage.set_shader_input("ColorLUT", lut)

    def load_grain(self):
        """ Loads the precomputed film grain """
        grain_tex = RPLoader.load_texture("/$$rp/data/film_grain/grain.txo")
        grain_tex.set_minfilter(SamplerState.FT_linear)
        grain_tex.set_magfilter(SamplerState.FT_linear)
        grain_tex.set_wrap_u(SamplerState.WM_repeat)
        grain_tex.set_wrap_v(SamplerState.WM_repeat)
        grain_tex.set_anisotropic_degree(0)
        self.stage.set_shader_input("PrecomputedGrain", grain_tex)

    def update_color_lut(self):
        self.debug("Updating color lut ..")
        self.load_lut()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec3
import direct.gui.DirectGuiGlobals as DGG

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.gui.checkbox import Checkbox
from metadrive.render_pipeline.rpcore.gui.text import Text


class LabeledCheckbox(RPObject):
    """ This is a checkbox, combined with a label. The arguments are
    equal to the Checkbox and OnscreenText arguments. """
    def __init__(
        self,
        parent=None,
        x=0,
        y=0,
        chb_callback=None,
        chb_args=None,
        chb_checked=True,
        text="",
        text_size=16,
        radio=False,
        text_color=None,
        expand_width=100,
        enabled=True
    ):
        """ Constructs a new checkbox, forwarding most of the elements to the
        underlying Checkbox and Text. """
        RPObject.__init__(self)
        if chb_args is None:
            chb_args = []

        if text_color is None:
            text_color = Vec3(1)

        if not enabled:
            text_color = Vec3(1.0, 0, 0.28)

        self.text_color = text_color

        self._checkbox = Checkbox(
            parent=parent,
            x=x,
            y=y,
            enabled=enabled,
            callback=chb_callback,
            extra_args=chb_args,
            checked=chb_checked,
            radio=radio,
            expand_width=expand_width
        )
        self._text = Text(
            x=x + 26,
            y=y + 9 + text_size // 4,
            text=text,
            align="left",
            parent=parent,
            size=text_size,
            color=text_color,
            may_change=True
        )

        if enabled:
            self._checkbox.node.bind(DGG.WITHIN, self._on_node_enter)
            self._checkbox.node.bind(DGG.WITHOUT, self._on_node_leave)

    def _on_node_enter(self, *args):  # pylint: disable=unused-argument
        """ Internal callback when the node gets hovered """
        self._text.node["fg"] = (self.text_color.x + 0.1, self.text_color.y + 0.1, self.text_color.z + 0.1, 1.0)

    def _on_node_leave(self, *args):  # pylint: disable=unused-argument
        """ Internal callback when the node gets no longer hovered """
        self._text.node["fg"] = (self.text_color.x, self.text_color.y, self.text_color.z, 1.0)

    @property
    def checkbox(self):
        """ Returns a handle to the checkbox """
        return self._checkbox

    @property
    def label(self):
        """ Returns a handle to the label """
        return self._text



import os

import matplotlib.pyplot as plt

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger
from metadrive.utils.draw_top_down_map import draw_top_down_map


def test_save_map_image():
    os.makedirs("tmp_images", exist_ok=True)
    setup_logger(debug=True)
    env = MetaDriveEnv(dict(num_scenarios=20, start_seed=0, map=10))
    try:
        for i in range(5):
            env.reset()
            surface = draw_top_down_map(env.current_map, resolution=(128, 128))
            plt.imshow(surface, cmap="Greys")
            plt.savefig("tmp_images/map_{}.png".format(i))
        env.close()
    finally:
        env.close()


if __name__ == "__main__":
    test_save_map_image()



import math
from shapely import geometry

from typing import Tuple, Union

import numpy as np

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.constants import DrivableAreaProperty
from metadrive.constants import PGLineType
from metadrive.utils.interpolating_line import InterpolatingLine
from metadrive.utils.math import get_points_bounding_box
from metadrive.utils.math import wrap_to_pi


class PointLane(AbstractLane, InterpolatingLine):
    """
    CenterLineLane is created by giving the center line points array or way points array.
    By using this lane type, map can be constructed from Waymo/nuPlan/OpenstreetMap dataset
    """
    VIS_LANE_WIDTH = 6.5
    POLYGON_SAMPLE_RATE = 1

    def __init__(
        self,
        center_line_points: Union[list, np.ndarray],
        width: float,
        polygon=None,
        forbidden: bool = False,
        speed_limit: float = 1000,
        priority: int = 0,
        need_lane_localization=True,
        auto_generate_polygon=True
    ):
        center_line_points = np.array(center_line_points)[..., :2]
        AbstractLane.__init__(self)
        InterpolatingLine.__init__(self, center_line_points)
        self._bounding_box = get_points_bounding_box(center_line_points)
        self._polygon = polygon
        self.width = width if width else self.VIS_LANE_WIDTH
        if self._polygon is None and auto_generate_polygon:
            self._polygon = self.auto_generate_polygon()
        self.shapely_polygon = geometry.Polygon(geometry.LineString(self._polygon))
        self.need_lane_localization = need_lane_localization
        self.set_speed_limit(speed_limit)
        self.forbidden = forbidden
        self.priority = priority
        # waymo lane line will be processed separately
        self.line_types = (PGLineType.NONE, PGLineType.NONE)
        self.is_straight = True if abs(self.heading_theta_at(0.1) -
                                       self.heading_theta_at(self.length - 0.1)) < np.deg2rad(10) else False
        self.start = self.position(0, 0)
        assert np.linalg.norm(self.start - center_line_points[0]) < 0.1, "Start point error!"
        self.end = self.position(self.length, 0)
        assert np.linalg.norm(self.end - center_line_points[-1]) < 1, "End point error!"

    def auto_generate_polygon(self):
        start_heading = self.heading_theta_at(0)
        start_dir = [math.cos(start_heading), math.sin(start_heading)]

        end_heading = self.heading_theta_at(self.length)
        end_dir = [math.cos(end_heading), math.sin(end_heading)]
        polygon = []
        longs = np.arange(0, self.length + self.POLYGON_SAMPLE_RATE, self.POLYGON_SAMPLE_RATE)
        for k in range(2):
            if k == 1:
                longs = longs[::-1]
            for t, longitude, in enumerate(longs):
                lateral = self.width_at(longitude) / 2
                lateral *= -1 if k == 0 else 1
                point = self.position(longitude, lateral)
                if (t == 0 and k == 0) or (t == len(longs) - 1 and k == 1):
                    # control the adding sequence
                    if k == 1:
                        # last point
                        polygon.append([point[0], point[1]])

                    # extend
                    polygon.append(
                        [
                            point[0] - start_dir[0] * self.POLYGON_SAMPLE_RATE,
                            point[1] - start_dir[1] * self.POLYGON_SAMPLE_RATE
                        ]
                    )

                    if k == 0:
                        # first point
                        polygon.append([point[0], point[1]])
                elif (t == 0 and k == 1) or (t == len(longs) - 1 and k == 0):

                    if k == 0:
                        # second point
                        polygon.append([point[0], point[1]])

                    polygon.append(
                        [
                            point[0] + end_dir[0] * self.POLYGON_SAMPLE_RATE,
                            point[1] + end_dir[1] * self.POLYGON_SAMPLE_RATE
                        ]
                    )

                    if k == 1:
                        # third point
                        polygon.append([point[0], point[1]])
                else:
                    polygon.append([point[0], point[1]])
        return np.asarray(polygon)

    def width_at(self, longitudinal: float) -> float:
        return self.width

    def heading_theta_at(self, longitudinal: float) -> float:
        """
        In rad
        """
        return self.get_heading_theta(longitudinal)

    def position(self, longitudinal: float, lateral: float) -> np.ndarray:
        return InterpolatingLine.position(self, longitudinal, lateral)

    def local_coordinates(self, position: Tuple[float, float], only_in_lane_point=False):
        return InterpolatingLine.local_coordinates(self, position, only_in_lane_point)

    def is_in_same_direction(self, another_lane):
        """
        Return True if two lane is in same direction
        """
        my_start_heading = self.heading_theta_at(0.1)
        another_start_heading = another_lane.heading_theta_at(0.1)

        my_end_heading = self.heading_theta_at(self.length - 0.1)
        another_end_heading = another_lane.heading_theta_at(self.length - 0.1)

        return True if abs(wrap_to_pi(my_end_heading) - wrap_to_pi(another_end_heading)) < 0.2 and abs(
            wrap_to_pi(my_start_heading) - wrap_to_pi(another_start_heading)
        ) < 0.2 else False

    def get_bounding_box(self):
        return self._bounding_box

    def destroy(self):
        self._bounding_box = None
        self.width = None
        self.forbidden = None
        self.priority = None

        self.line_types = None
        self.is_straight = None
        self.start = None
        self.end = None
        self._polygon = None
        InterpolatingLine.destroy(self)
        super(PointLane, self).destroy()

    def construct_lane_in_block(self, block, lane_index):
        """
        Modified from base class, the width is set to 6.5
        """
        if lane_index is not None:
            self.index = lane_index
        assert self._polygon is not None, "Polygon is required for building lane"
        # build visualization
        segment_num = int(self.length / DrivableAreaProperty.LANE_SEGMENT_LENGTH)
        if segment_num == 0:
            middle = self.position(self.length / 2, 0)
            end = self.position(self.length, 0)
            theta = self.heading_theta_at(self.length / 2)
            self._construct_lane_only_vis_segment(block, middle, self.VIS_LANE_WIDTH, self.length, theta)

        for i in range(segment_num):
            middle = self.position(self.length * (i + .5) / segment_num, 0)
            end = self.position(self.length * (i + 1) / segment_num, 0)
            direction_v = end - middle
            theta = math.atan2(direction_v[1], direction_v[0])
            length = self.length
            self._construct_lane_only_vis_segment(block, middle, self.VIS_LANE_WIDTH, length * 1.3 / segment_num, theta)

        # build physics contact
        if self.need_lane_localization:
            self._construct_lane_only_physics_polygon(block, self._polygon)

    @property
    def polygon(self):
        return self._polygon

    def point_on_lane(self, point):
        s_point = geometry.Point(point[0], point[1])
        return self.shapely_polygon.contains(s_point)



_metadrive_class_list = None
_metadrive_class_registry = None


def _initialize_registry():
    global _metadrive_class_list
    _metadrive_class_list = []

    # Register all PG blocks
    from metadrive.component.pgblock.bottleneck import Merge, Split
    from metadrive.component.pgblock.curve import Curve
    from metadrive.component.pgblock.fork import InFork, OutFork
    from metadrive.component.pgblock.parking_lot import ParkingLot
    from metadrive.component.pgblock.ramp import InRampOnStraight, OutRampOnStraight
    from metadrive.component.pgblock.roundabout import Roundabout
    from metadrive.component.pgblock.std_intersection import StdInterSection
    from metadrive.component.pgblock.std_t_intersection import StdTInterSection
    from metadrive.component.pgblock.straight import Straight
    from metadrive.component.pgblock.tollgate import TollGate
    from metadrive.component.pgblock.bidirection import Bidirection
    _metadrive_class_list.extend(
        [
            Merge, Split, Curve, InFork, OutFork, ParkingLot, InRampOnStraight, OutRampOnStraight, Roundabout,
            StdInterSection, StdTInterSection, Straight, TollGate, Bidirection
        ]
    )

    global _metadrive_class_registry
    _metadrive_class_registry = {k.__name__: k for k in _metadrive_class_list}


def get_metadrive_class(class_name):
    global _metadrive_class_registry
    if _metadrive_class_registry is None:
        _initialize_registry()

    assert class_name in _metadrive_class_registry, "{} is not in Registry: {}".format(
        class_name, _metadrive_class_registry.keys()
    )
    return _metadrive_class_registry[class_name]



from metadrive.envs import MetaDriveEnv


# Related issue:
# https://github.com/metadriverse/metadrive/issues/191
def test_close_and_reset():

    env = MetaDriveEnv({"start_seed": 1000, "num_scenarios": 1})
    eval_env = MetaDriveEnv()
    assert eval_env.action_space.contains(env.action_space.sample())
    env.reset()
    for i in range(100):
        env.step(env.action_space.sample())
    env.close()
    env.reset()
    for i in range(100):
        env.step(env.action_space.sample())
    env.close()

    eval_env.reset()
    for i in range(100):
        eval_env.step(eval_env.action_space.sample())
    eval_env.close()
    env.reset()
    for i in range(100):
        env.step(env.action_space.sample())
    env.close()


if __name__ == '__main__':
    test_close_and_reset()



from metadrive.policy.manual_control_policy import ManualControlPolicy


class AIProtectPolicy(ManualControlPolicy):
    """
    This policy can protect Manual control and EnvInputControl
    """
    def act(self, agent_id):
        action = super(AIProtectPolicy, self).act(agent_id)
        vehicle = self.engine.agent_manager.active_agents[agent_id]
        steering = action[0]
        throttle = action[1]
        if not vehicle.expert_takeover:
            # saver can be used for human or another AI
            save_level = self.engine.global_config["save_level"] if not vehicle.expert_takeover else 1.0
            from metadrive.examples.ppo_expert import expert
            try:
                saver_a, obs = expert(vehicle, deterministic=False, need_obs=True)
                obs = obs[0]
            except ValueError:
                print("Expert can not takeover, due to observation space mismathing!")
                saver_a = action
            else:
                if save_level > 0.9:
                    steering = saver_a[0]
                    throttle = saver_a[1]
                elif save_level > 1e-3:
                    heading_diff = vehicle.heading_diff(vehicle.lane) - 0.5
                    f = min(1 + abs(heading_diff) * vehicle.speed_km_h * vehicle.max_speed_km_h, save_level * 10)
                    # for out of road
                    if (obs[0] < 0.04 * f and heading_diff < 0) or (obs[1] < 0.04 * f and heading_diff > 0) or obs[
                        0] <= 1e-3 or \
                            obs[
                                1] <= 1e-3:
                        steering = saver_a[0]
                        throttle = saver_a[1]
                        if vehicle.speed_km_h < 5:
                            throttle = 0.5
                    # for collision
                    lidar_p = self.engine.agent_manager.observations[vehicle.id].cloud_points
                    left = int(vehicle.lidar.num_lasers / 4)
                    right = int(vehicle.lidar.num_lasers / 4 * 3)
                    if min(lidar_p[left - 4:left + 6]) < (save_level + 0.1) / 10 or min(lidar_p[right - 4:right + 6]
                                                                                        ) < (save_level + 0.1) / 10:
                        # lateral safe distance 2.0m
                        steering = saver_a[0]
                    if action[1] >= 0 and saver_a[1] <= 0 and min(min(lidar_p[0:10]), min(lidar_p[-10:])) < save_level:
                        # longitude safe distance 15 m
                        throttle = saver_a[1]

        # indicate if current frame is takeover step
        pre_save = vehicle.takeover
        vehicle.takeover = True if action[0] != steering or action[1] != throttle else False
        self.action_info = {
            "takeover_start": True if not pre_save and vehicle.takeover else False,
            "takeover_end": True if pre_save and not vehicle.takeover else False,
            "takeover": vehicle.takeover if pre_save else False
        }
        action = (steering, throttle) if self.action_info["takeover"] else action
        self.action_info["action"] = action
        return action



import argparse
import time

import cv2
import torch
from torch.utils.dlpack import from_dlpack

from metadrive import MetaDriveEnv
from metadrive.policy.idm_policy import IDMPolicy


def _test_rgb_camera_as_obs(render=False, image_on_cuda=True, debug=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            agent_policy=IDMPolicy,
            traffic_density=0.0,
            image_observation=True,
            image_on_cuda=True if image_on_cuda else False,
            use_render=False,
            vehicle_config=dict(image_source="main_camera"),
            show_interface=False,
            show_logo=False,
            show_fps=False,
        )
    )
    o, i = env.reset()

    # for debug
    if debug:
        ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
        cv2.imwrite("reset_frame.png", ret * 255)

    action = [0.0, 0.1]
    start = time.time()
    print(
        "Use {} with resolution {}".format(
            env.config["vehicle_config"]["image_source"],
            (env.observation_space["image"].shape[0], env.observation_space["image"].shape[1])
        )
    )
    for i in range(20000):
        o, r, d, _, _ = env.step(action)
        if image_on_cuda:
            torch_tensor = from_dlpack(o["image"].toDlpack())
        else:
            torch_tensor = torch.Tensor(o["image"])

        if debug:
            cv2.imwrite(
                "{}_frame.png".format(i),
                (o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]) * 255
            )

        if render:
            ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
            cv2.imshow("window", ret)
            cv2.waitKey(1)
        if d:
            print("FPS: {}".format(i / (time.time() - start)))
            break


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--render", action="store_true")
    parser.add_argument("--native", action="store_true")
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()
    _test_rgb_camera_as_obs(args.render, image_on_cuda=not args.native, debug=args.debug)
    print("Test Successful !! The FPS should go beyond 400 FPS if you are running on GPUs better than RTX 3060.")



#!/usr/bin/env python3
# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import json
import struct


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('input', type=str, help='Input .gltf file path.')
    parser.add_argument('--extras', action='store_true', required=False, help="Shows node's extras.")

    return parser.parse_args()


def print_node(gltf_data, node_id, joints=None, skeletons=None, indent=1, parent_node=None, extras=None):
    gltf_node = gltf_data['nodes'][node_id]

    type_ = 'N'
    extra = ''

    matrix = ''
    if 'translation' in gltf_node:
        if sum(gltf_node['translation']) != 0:
            matrix += 'T'
    if 'rotation' in gltf_node:
        if not (gltf_node['rotation'][0] == 0 and gltf_node['rotation'][1] == 0 and gltf_node['rotation'][2] == 0
                and gltf_node['rotation'][3] == 1):
            matrix += 'R'
    if 'scale' in gltf_node:
        if not (gltf_node['scale'][0] == 1 and gltf_node['scale'][1] == 1 and gltf_node['scale'][2] == 1):
            matrix += 'S'
    if 'matrix' in gltf_node:
        matrix += 'M'

    if matrix:
        extra += ' <{}>'.format(matrix)

    if node_id in (skeletons or []):
        type_ = 'S'  # skeleton/armature
    elif node_id in (joints or []):
        type_ = 'J'  # joint/bone

    if 'mesh' in gltf_node:
        type_ = 'M'  # mesh/geometry

    if 'skin' in gltf_node:
        refs = []

        if 'skin' in gltf_node:
            skin_id = gltf_node['skin']
            gltf_skin = gltf_data['skins'][skin_id]
            v = '{} ({} joints)'.format(gltf_skin.get('name', 'SKIN #{}'.format(skin_id)), len(gltf_skin['joints']))
            refs.append(('skin', v))

            if 'skeleton' in gltf_skin:
                skeleton_id = gltf_skin['skeleton']
                gltf_skeleton = gltf_data['nodes'][skeleton_id]
                v = '{}'.format(gltf_skeleton.get('name', 'SKELETON #{}'.format(skeleton_id)))
                refs.append(('skeleton', v))

        if 'mesh' in gltf_node:
            mesh_id = gltf_node['mesh']
            gltf_mesh = gltf_data['meshes'][mesh_id]
            refs.append(('mesh', gltf_mesh['name']))

        extra += ' {' + ', '.join(['{}: {}'.format(*i) for i in refs]) + '}'

    if 'VRM' in (gltf_data.get('extensions') or {}):
        vrm_extra = []

        vrm_bones = gltf_data['extensions']['VRM']['humanoid']['humanBones']
        for vrm_bone in vrm_bones:
            if node_id == vrm_bone['node']:
                vrm_extra.append('VRM bone: {}'.format(vrm_bone['bone']))

        for group in gltf_data['extensions']['VRM']['secondaryAnimation']['boneGroups']:
            if node_id in group['bones']:
                vrm_extra.append('bonegroup')
                break

        if vrm_extra:
            extra += ' {%s}' % ', '.join(vrm_extra)

    # for child_node_id in gltf_node.get('children', []):
    #     child_gltf_node = gltf_data['nodes'][child_node_id]
    #     if 'skin' in child_gltf_node:
    #         # type_ = 'S'  # skeleton/armature
    #         skin_id = child_gltf_node['skin']
    #         gltf_skin = gltf_data['skins'][skin_id]
    #         # joints = gltf_skin['joints']

    is_ = ''
    for i in range(indent):
        if i < indent - 1:
            is_ += '  |'
        else:
            is_ += '  +'
    print('{} [{}] {}{}'.format(is_, type_, gltf_node['name'], extra))

    if extras:
        for k, v in gltf_node.get('extras', {}).items():
            print('   {}  {}: {}'.format(is_, k, v))

    for child_node_id in gltf_node.get('children', []):
        print_node(
            gltf_data, child_node_id, joints=joints, skeletons=skeletons, indent=indent + 1, parent_node=gltf_node
        )


def print_scene(gltf_data, scene_id, extras=False):
    gltf_scene = gltf_data['scenes'][scene_id]
    print(' [R] {}'.format(gltf_scene.get('name', 'SCENE')))

    # child to parent mapping
    parents = {}
    for parent_id, gltf_node in enumerate(gltf_data['nodes']):
        for child_id in gltf_node.get('children', ()):
            parents[child_id] = parent_id

    skeletons = set()
    joints = set()
    for gltf_skin in gltf_data['skins']:
        joints |= set(gltf_skin['joints'])
        # search for the root bone
        for joint_id in gltf_skin['joints']:
            if parents.get(joint_id) not in gltf_skin['joints']:  # no parent bone
                skeletons.add(joint_id)  # mark root bone as skeleton
                break

    for node_id in gltf_scene['nodes']:
        print_node(gltf_data, node_id, joints=joints, skeletons=skeletons, extras=extras)


def print_anim(gltf_data, gltf_anim):
    extra = ''
    input_id = gltf_anim['samplers'][0]['input']
    extra += '{} frames'.format(gltf_data['accessors'][input_id]['count'])
    if extra:
        extra = ' {' + extra + '}'

    print(' [A] {}{}'.format(gltf_anim['name'], extra))


def print_mat(gltf_data, gltf_mat):
    tex_ids = []
    if 'baseColorTexture' in gltf_mat.get('pbrMetallicRoughness', {}):
        tex_ids.append(('Color', gltf_mat['pbrMetallicRoughness']['baseColorTexture']['index']))
    if 'metallicRoughnessTexture' in gltf_mat.get('pbrMetallicRoughness', {}):
        tex_ids.append(('MetRough', gltf_mat['pbrMetallicRoughness']['metallicRoughnessTexture']['index']))
    if 'normalTexture' in gltf_mat:
        tex_ids.append(('Norm', gltf_mat['normalTexture']['index']))
    if 'emissiveTexture' in gltf_mat:
        tex_ids.append(('Emit', gltf_mat['emissiveTexture']['index']))

    print(' [M] {}'.format(gltf_mat['name']))
    for tex_type, tex_id in tex_ids:
        print_tex(gltf_data, tex_type, gltf_data['textures'][tex_id])


def print_tex(gltf_data, gltf_tex_type, gltf_tex):
    sampler = gltf_data['samplers'][gltf_tex['sampler']]
    # source = gltf_data['images'][gltf_tex['source']]
    print('  + [T] {name} <{type}>'.format(**{
        'type': gltf_tex_type,
        'name': sampler.get('name', 'SAMPLER'),
    }))


def main():
    args = parse_args()
    gltf_data = {
        'scene': {},
    }

    if args.input.endswith('.glb') or args.input.endswith('.vrm'):
        with open(args.input, 'rb') as f:
            assert f.read(4) == b'glTF'  # header
            assert struct.unpack('<I', f.read(4))[0] == 2  # version
            full_size = struct.unpack('<I', f.read(4))

            chunk_type = None
            chunk_data = None
            while True:
                chunk_size = struct.unpack('<I', f.read(4))[0]
                chunk_type = f.read(4)
                chunk_data = f.read(chunk_size)
                if chunk_type == b'JSON':
                    break

            if chunk_type == b'JSON':
                gltf_data = json.loads(chunk_data)

    else:
        with open(args.input, 'r') as f:
            gltf_data = json.load(f)

    if 'generator' in gltf_data.get('asset', {}):
        print('glTF generator: {}'.format(gltf_data['asset']['generator']))

    if 'VRM' in gltf_data.get('extensions', {}):
        vrm_meta = gltf_data['extensions']['VRM']
        print('VRM exporter: {}'.format(vrm_meta['exporterVersion']))

    print_scene(gltf_data, gltf_data['scene'], extras=args.extras)

    for gltf_anim in (gltf_data.get('animations') or []):
        print_anim(gltf_data, gltf_anim)

    for gltf_mat in (gltf_data.get('materials') or []):
        print_mat(gltf_data, gltf_mat)


if __name__ == '__main__':
    main()



import logging
from metadrive.version import VERSION
import time
from collections import defaultdict
from typing import Union, Dict, AnyStr, Optional, Tuple, Callable

import gymnasium as gym
import numpy as np
from panda3d.core import PNMImage

from metadrive.component.sensors.base_camera import BaseCamera
from metadrive.component.sensors.vehicle_panel import VehiclePanel
from metadrive.constants import RENDER_MODE_NONE, DEFAULT_AGENT
from metadrive.constants import RENDER_MODE_ONSCREEN, RENDER_MODE_OFFSCREEN
from metadrive.constants import TerminationState
from metadrive.engine.engine_utils import initialize_engine, close_engine, \
    engine_initialized, set_global_random_seed, initialize_global_config, get_global_config
from metadrive.engine.logger import get_logger
from metadrive.manager.agent_manager import AgentManager
from metadrive.manager.record_manager import RecordManager
from metadrive.manager.replay_manager import ReplayManager
from metadrive.obs.image_obs import ImageStateObservation
from metadrive.obs.observation_base import ObservationBase
from metadrive.obs.state_obs import LidarStateObservation
from metadrive.policy.env_input_policy import EnvInputPolicy
from metadrive.scenario.utils import convert_recorded_scenario_exported
from metadrive.utils import Config, merge_dicts, get_np_random, concat_step_infos

BASE_DEFAULT_CONFIG = dict(

    # ===== agent =====
    random_agent_model=False,
    agent_policy=EnvInputPolicy,

    # ===== multi-agent =====
    num_agents=1,  # Note that this can be set to >1 in MARL envs, or set to -1 for as many vehicles as possible.
    is_multi_agent=False,
    allow_respawn=False,
    delay_done=0,  # How many steps for the agent to stay static at the death place after done.
    # Whether only return single agent-like observation and action space
    return_single_space=False,

    # ===== Action =====
    manual_control=False,
    controller="keyboard",  # "joystick" or "keyboard"
    decision_repeat=5,
    discrete_action=False,
    discrete_steering_dim=5,
    discrete_throttle_dim=5,
    # When discrete_action=True: If True, use MultiDiscrete action space. Otherwise, use Discrete.
    use_multi_discrete=False,
    action_check=False,

    # ===== Observation =====
    rgb_clip=True,  # clip rgb to (0, 1)
    rgb_to_grayscale=False,  # whether to convert rgb image to grayscale in class ImageObservation
    stack_size=3,  # the number of timesteps for stacking image observation
    image_observation=False,  # use image observation or lidar

    # ===== Rendering =====
    use_render=False,  # if true pop a window to render
    debug=False,
    disable_model_compression=False,  # disable compression if you wish to launch the window quicker.
    cull_scene=True,  # only for debug use
    use_chase_camera_follow_lane=False,  # If true, then vision would be more stable.
    camera_height=2.2,
    camera_dist=7.5,
    camera_pitch=None,  # degree
    camera_smooth=True,  # degree
    camera_fov=65,
    prefer_track_agent=None,
    draw_map_resolution=1024,  # Drawing the map in a canvas of (x, x) pixels.
    top_down_camera_initial_x=0,
    top_down_camera_initial_y=0,
    top_down_camera_initial_z=200,  # height
    show_logo=True,

    # ===== Vehicle =====
    vehicle_config=dict(
        increment_steering=False,
        vehicle_model="default",
        show_navi_mark=True,
        extra_action_dim=0,
        enable_reverse=False,
        random_navi_mark_color=False,
        show_dest_mark=False,
        show_line_to_dest=False,
        show_line_to_navi_mark=False,
        use_special_color=False,
        no_wheel_friction=False,

        # ===== use image =====
        image_source="rgb_camera",  # take effect when only when image_observation == True

        # ===== vehicle spawn and destination =====
        navigation_module=None,  # a class type for self-defined navigation
        need_navigation=True,
        spawn_lane_index=None,
        spawn_longitude=5.0,
        spawn_lateral=0.0,
        destination=None,
        spawn_position_heading=None,
        spawn_velocity=None,  # m/s
        spawn_velocity_car_frame=False,

        # ==== others ====
        overtake_stat=False,  # we usually set to True when evaluation
        random_color=False,
        # The shape of vehicle are predefined by its class. But in special scenario (WaymoVehicle) we might want to
        # set to arbitrary shape.
        width=None,
        length=None,
        height=None,
        mass=None,

        # ===== vehicle module config =====
        lidar=dict(
            num_lasers=240, distance=50, num_others=0, gaussian_noise=0.0, dropout_prob=0.0, add_others_navi=False
        ),
        side_detector=dict(num_lasers=0, distance=50, gaussian_noise=0.0, dropout_prob=0.0),
        lane_line_detector=dict(num_lasers=0, distance=20, gaussian_noise=0.0, dropout_prob=0.0),
        show_lidar=False,
        show_side_detector=False,
        show_lane_line_detector=False,
        gaussian_noise=0.0,
        dropout_prob=0.0,
        light=False,  # vehicle light, only available when enabling render-pipeline
    ),

    # ===== Sensors =====
    # It should be a dict like this:
    # sensors={
    #   "rgb_camera": (RGBCamera, arg_1, arg_2, ..., arg_n),
    #    ...
    #   "your_sensor_name": (sensor_class, arg_1, arg_2, ..., arg_n)
    # }
    # Example:
    # sensors = dict(
    #         # rgb_camera=(RGBCamera, 84, 84),
    #         # mini_map=(MiniMap, 84, 84, 250),
    #         # depth_camera=(DepthCamera, 84, 84),
    #         )
    # These sensors will be constructed automatically and can be accessed in engine.get_sensor("sensor_name")
    # NOTE: main_camera will be added automatically if you are using offscreen/onscreen mode
    sensors={},

    # when main_camera is not the image_source for vehicle, reduce the window size to (1,1) for boosting efficiency
    auto_resize_window=True,

    # ===== Agent config =====
    target_vehicle_configs={DEFAULT_AGENT: dict(use_special_color=False, spawn_lane_index=None)},

    # ===== Engine Core config =====
    window_size=(1200, 900),  # or (width, height), if set to None, it will be automatically determined
    physics_world_step_size=2e-2,
    show_fps=True,
    global_light=True,
    # only render physics world without model, a special debug option
    debug_physics_world=False,
    # debug static world
    debug_static_world=False,
    # (Deprecated) set to true only when on headless machine and use rgb image!!!!!!
    # turn on to profile the efficiency
    pstats=False,
    # this is an advanced feature for accessing image with moving them to ram!
    image_on_cuda=False,
    # We will determine the render mode automatically, it runs at physics-only mode by default
    _render_mode=RENDER_MODE_NONE,
    # accelerate the lidar perception
    _disable_detector_mask=False,
    # None: unlimited, number: fps
    force_render_fps=None,
    # if set to True all objects will be force destroyed when call clear()
    force_destroy=False,
    # number of buffering objects for each class.
    # we will maintain a set of buffers in the engine to store the used objects and can reuse them
    # when possible. But it is possible that some classes of objects are always forcefully respawn
    # and thus those used objects are stored in the buffer and never be reused.
    num_buffering_objects=200,
    # turn on to use render pipeline, which provides advanced rendering effects (Beta)
    render_pipeline=False,
    # daytime is only available when using render-pipeline
    daytime="19:00",  # use string like "13:40", We usually set this by editor in toolkit
    # debug panda3d
    debug_panda3d=False,

    # ===== Mesh Terrain =====
    # road will have a marin whose width is determined by this value, unit: [m]
    drivable_region_extension=6,
    # height scale for mountains, unit: [m]
    height_scale=120,

    # ===== Others =====
    # Force to generate objects in the left lane.
    _debug_crash_object=False,
    horizon=None,  # The maximum length of each environmental episode. Set to None to remove this constraint
    max_step_per_agent=None,  # The maximum length of each agent episode. Raise max_step termination when reaches.
    show_interface_navi_mark=True,
    show_mouse=True,
    show_skybox=True,
    show_terrain=True,
    show_interface=True,
    show_policy_mark=False,  # show marks for policies for debugging multi-policy setting
    show_coordinates=False,  # show coordinates for maps and objects for debug
    interface_panel=["dashboard"],  # a list whose element chosen from sensors.keys() + "dashboard"
    multi_thread_render=True,
    multi_thread_render_mode="Cull",  # or "Cull/Draw"
    preload_models=True,  # preload pedestrian Object for avoiding lagging when creating it for the first time
    log_level=logging.INFO,

    # record/replay metadata
    record_episode=False,  # when replay_episode is not None ,this option will be useless
    replay_episode=None,  # set the replay file to enable replay
    only_reset_when_replay=False,  # Scenario will only be initialized, while future trajectories will not be replayed
    force_reuse_object_name=False,  # If True, when restoring objects, use the same ID as in dataset
)


class BaseEnv(gym.Env):
    # Force to use this seed if necessary. Note that the recipient of the forced seed should be explicitly implemented.
    _DEBUG_RANDOM_SEED: Union[int, None] = None
    DEFAULT_AGENT = DEFAULT_AGENT

    @classmethod
    def default_config(cls) -> Config:
        return Config(BASE_DEFAULT_CONFIG)

    # ===== Intialization =====
    def __init__(self, config: dict = None):
        if config is None:
            config = {}
        self.logger = get_logger(
            self.logger_name, config.get("log_level", logging.DEBUG if config.get("debug", False) else logging.INFO)
        )
        merged_config = self.default_config().update(config, False, ["target_vehicle_configs", "sensors"])
        global_config = self._post_process_config(merged_config)
        global_config["vehicle_config"]["main_camera"] = global_config["window_size"]

        self.config = global_config
        initialize_global_config(self.config)

        # agent check
        self.num_agents = self.config["num_agents"]
        self.is_multi_agent = self.config["is_multi_agent"]
        if not self.is_multi_agent:
            assert self.num_agents == 1
        else:
            assert not self.config["image_on_cuda"], "Image on cuda don't support Multi-agent!"
        assert isinstance(self.num_agents, int) and (self.num_agents > 0 or self.num_agents == -1)

        # observation and action space
        self.agent_manager = self._get_agent_manager()

        # lazy initialization, create the main vehicle in the lazy_init() func
        # self.engine: Optional[BaseEngine] = None

        # In MARL envs with respawn mechanism, varying episode lengths might happen.
        self.episode_rewards = defaultdict(float)
        self.episode_lengths = defaultdict(int)

    def _post_process_config(self, config):
        """Add more special process to merged config"""
        # Cancel interface panel
        self.logger.info("MetaDrive version: {}".format(VERSION))
        if not config["show_interface"]:
            config["interface_panel"] = []

        # Optimize main window
        if not config["use_render"] and config["image_observation"] and \
                config["vehicle_config"]["image_source"] != "main_camera" and config["auto_resize_window"]:
            # reduce size as we don't use the main camera content for improving efficiency
            config["window_size"] = (1, 1)
            self.logger.debug(
                "Main window size is reduced to (1, 1) for boosting efficiency."
                "To cancel this, set auto_resize_window = False"
            )

        # Optimize sensor creation in none-screen mode
        if not config["use_render"] and not config["image_observation"]:
            filtered = {}
            for id, cfg in config["sensors"].items():
                if not issubclass(cfg[0], BaseCamera):
                    filtered[id] = cfg
            config["sensors"] = filtered
            config["interface_panel"] = []

        # Merge vehicle_panel config with sensors
        to_use = []
        for panel in config["interface_panel"]:
            if panel == "dashboard":
                config["sensors"]["dashboard"] = (VehiclePanel, )
            if panel not in config["sensors"]:
                self.logger.warning("Fail to add sensor: {} to the interface. Remove it from panel list!".format(panel))
            else:
                to_use.append(panel)
        config["interface_panel"] = to_use

        # Check sensor existence
        if config["use_render"] or config["image_observation"]:
            config["sensors"]["main_camera"] = ("MainCamera", *config["window_size"])
        if config["image_observation"]:
            assert config["vehicle_config"]["image_source"] in config["sensors"], \
                "Can not find sensor with id: {} from existing sensors: {}".format(
                    config["vehicle_config"]["image_source"], config["sensors"])

        # show sensor lists
        _str = "Sensors: [{}]"
        sensors_str = ""
        for _id, cfg in config["sensors"].items():
            sensors_str += "{}: {}{}, ".format(_id, cfg[0] if isinstance(cfg[0], str) else cfg[0].__name__, cfg[1:])
        self.logger.info(_str.format(sensors_str[:-2]))

        # determine render mode automatically
        if config["use_render"]:
            assert "main_camera" in config["sensors"]
            config["_render_mode"] = RENDER_MODE_ONSCREEN
        else:
            config["_render_mode"] = RENDER_MODE_NONE
            for sensor in config["sensors"].values():
                if sensor[0] == "MainCamera" or (issubclass(BaseCamera, sensor[0]) and sensor[0] != VehiclePanel):
                    config["_render_mode"] = RENDER_MODE_OFFSCREEN
                    break
        self.logger.info("Render Mode: {}".format(config["_render_mode"]))
        return config

    def _get_observations(self) -> Dict[str, "ObservationBase"]:
        raise NotImplementedError()

    def _get_observation_space(self):
        return {v_id: obs.observation_space for v_id, obs in self.observations.items()}

    def _get_action_space(self):
        if self.is_multi_agent:
            return {
                v_id: self.config["agent_policy"].get_input_space()
                for v_id in self.config["target_vehicle_configs"].keys()
            }
        else:
            return {DEFAULT_AGENT: self.config["agent_policy"].get_input_space()}

    def _get_agent_manager(self):
        return AgentManager(init_observations=self._get_observations(), init_action_space=self._get_action_space())

    def lazy_init(self):
        """
        Only init once in runtime, variable here exists till the close_env is called
        :return: None
        """
        # It is the true init() func to create the main vehicle and its module, to avoid incompatible with ray
        if engine_initialized():
            return
        initialize_engine(self.config)
        # engine setup
        self.setup_engine()
        # other optional initialization
        self._after_lazy_init()

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()

    def _after_lazy_init(self):
        pass

    # ===== Run-time =====
    def step(self, actions: Union[np.ndarray, Dict[AnyStr, np.ndarray], int]):
        actions = self._preprocess_actions(actions)
        engine_info = self._step_simulator(actions)
        return self._get_step_return(actions, engine_info=engine_info)

    def _preprocess_actions(self, actions: Union[np.ndarray, Dict[AnyStr, np.ndarray], int]) \
            -> Union[np.ndarray, Dict[AnyStr, np.ndarray], int]:
        if not self.is_multi_agent:
            actions = {v_id: actions for v_id in self.vehicles.keys()}
        else:
            if self.config["action_check"]:
                # Check whether some actions are not provided.
                given_keys = set(actions.keys())
                have_keys = set(self.vehicles.keys())
                assert given_keys == have_keys, "The input actions: {} have incompatible keys with existing {}!".format(
                    given_keys, have_keys
                )
            else:
                # That would be OK if extra actions is given. This is because, when evaluate a policy with naive
                # implementation, the "termination observation" will still be given in T=t-1. And at T=t, when you
                # collect action from policy(last_obs) without masking, then the action for "termination observation"
                # will still be computed. We just filter it out here.
                actions = {v_id: actions[v_id] for v_id in self.vehicles.keys()}
        return actions

    def _step_simulator(self, actions):
        # Note that we use shallow update for info dict in this function! This will accelerate system.
        scene_manager_before_step_infos = self.engine.before_step(actions)
        # step all entities
        self.engine.step(self.config["decision_repeat"])
        # update states, if restore from episode data, position and heading will be force set in update_state() function
        scene_manager_after_step_infos = self.engine.after_step()
        return merge_dicts(
            scene_manager_after_step_infos, scene_manager_before_step_infos, allow_new_keys=True, without_copy=True
        )

    def reward_function(self, vehicle_id: str) -> Tuple[float, Dict]:
        """
        Override this func to get a new reward function
        :param vehicle_id: name of this base vehicle
        :return: reward, reward info
        """
        raise NotImplementedError()

    def cost_function(self, vehicle_id: str) -> Tuple[float, Dict]:
        raise NotImplementedError()

    def done_function(self, vehicle_id: str) -> Tuple[bool, Dict]:
        raise NotImplementedError()

    def render(self,
               text: Optional[Union[dict, str]] = None,
               return_bytes=False,
               mode=None,
               *args,
               **kwargs) -> Optional[np.ndarray]:
        """
        This is a pseudo-render function, only used to update onscreen message when using panda3d backend
        :param text:text to show
        :return: when mode is 'rgb', image array is returned
        """

        if mode in ["top_down", "topdown", "bev", "birdview"]:
            ret = self._render_topdown(text=text, *args, **kwargs)
            return ret
        assert self.config["use_render"] or self.engine.mode != RENDER_MODE_NONE, \
            ("Panda Rendering is off now, can not render. Please set config['use_render'] = True!")

        self.engine.render_frame(text)

        if mode != "human" and self.config["image_observation"]:
            # fetch img from img stack to be make this func compatible with other render func in RL setting
            return self.observations[DEFAULT_AGENT].img_obs.get_image()

        if mode == "rgb_array":
            assert self.config["use_render"], "You should create a Panda3d window before rendering images!"
            # if not hasattr(self, "temporary_img_obs"):
            #     from metadrive.obs.image_obs import ImageObservation
            #     image_source = "rgb_camera"
            #     assert len(self.vehicles) == 1, "Multi-agent not supported yet!"
            #     self.temporary_img_obs = ImageObservation(self.vehicles[DEFAULT_AGENT].config, image_source, False)
            # # else:
            # #     raise ValueError("Not implemented yet!")
            # self.temporary_img_obs.observe(self.vehicles[DEFAULT_AGENT])
            # return self.temporary_img_obs.get_image()
            return self.engine.get_window_image(return_bytes=return_bytes)
        return None

    def reset(self, seed: Union[None, int] = None):
        """
        Reset the env, scene can be restored and replayed by giving episode_data
        Reset the environment or load an episode from episode data to recover is
        :param seed: The seed to set the env.
        :return: None
        """
        if self.logger is None:
            self.logger = get_logger(
                self.logger_name,
                self.config.get("log_level", logging.DEBUG if self.config.get("debug", False) else logging.INFO)
            )
        self.lazy_init()  # it only works the first time when reset() is called to avoid the error when render
        self._reset_global_seed(seed)
        if self.engine is None:
            raise ValueError(
                "Current MetaDrive instance is broken. Please make sure there is only one active MetaDrive "
                "environment exists in one process. You can try to call env.close() and then call "
                "env.reset() to rescue this environment. However, a better and safer solution is to check the "
                "singleton of MetaDrive and restart your program."
            )
        self.engine.reset()
        if self._top_down_renderer is not None:
            self._top_down_renderer.reset(self.current_map)

        self.dones = {agent_id: False for agent_id in self.vehicles.keys()}
        self.episode_rewards = defaultdict(float)
        self.episode_lengths = defaultdict(int)

        assert (len(self.vehicles) == self.num_agents) or (self.num_agents == -1)
        assert self.config is self.engine.global_config is get_global_config(), "Inconsistent config may bring errors!"
        return self._get_reset_return()

    def _get_reset_return(self):
        # TODO: figure out how to get the information of the before step
        scene_manager_before_step_infos = {}
        scene_manager_after_step_infos = self.engine.after_step()

        obses = {}
        done_infos = {}
        cost_infos = {}
        reward_infos = {}
        engine_info = merge_dicts(
            scene_manager_after_step_infos, scene_manager_before_step_infos, allow_new_keys=True, without_copy=True
        )
        for v_id, v in self.vehicles.items():
            self.observations[v_id].reset(self, v)
            obses[v_id] = self.observations[v_id].observe(v)
            _, reward_infos[v_id] = self.reward_function(v_id)
            _, done_infos[v_id] = self.done_function(v_id)
            _, cost_infos[v_id] = self.cost_function(v_id)

        step_infos = concat_step_infos([engine_info, done_infos, reward_infos, cost_infos])

        if self.is_multi_agent:
            return (obses, step_infos)
        else:
            return (self._wrap_as_single_agent(obses), self._wrap_as_single_agent(step_infos))

    def _get_step_return(self, actions, engine_info):
        # update obs, dones, rewards, costs, calculate done at first !
        obses = {}
        done_infos = {}
        cost_infos = {}
        reward_infos = {}
        rewards = {}
        for v_id, v in self.vehicles.items():
            rewards[v_id], reward_infos[v_id] = self.reward_function(v_id)
            done_function_result, done_infos[v_id] = self.done_function(v_id)
            _, cost_infos[v_id] = self.cost_function(v_id)
            done = done_function_result or self.dones[v_id]
            self.dones[v_id] = done
            o = self.observations[v_id].observe(v)
            obses[v_id] = o

        step_infos = concat_step_infos([engine_info, done_infos, reward_infos, cost_infos])

        # For extreme scenario only. Force to terminate all vehicles if the environmental step exceeds 5 times horizon.
        should_external_done = False
        if self.config["horizon"] is not None:
            should_external_done = self.episode_step > 5 * self.config["horizon"]
        if should_external_done:
            for k in self.dones:
                self.dones[k] = True

        terminateds = {k: self.dones[k] for k in self.vehicles.keys()}
        truncateds = {k: False for k in self.vehicles.keys()}

        for v_id, r in rewards.items():
            self.episode_rewards[v_id] += r
            step_infos[v_id]["episode_reward"] = self.episode_rewards[v_id]
            self.episode_lengths[v_id] += 1
            step_infos[v_id]["episode_length"] = self.episode_lengths[v_id]

        if not self.is_multi_agent:
            return self._wrap_as_single_agent(obses), self._wrap_as_single_agent(rewards), \
                   self._wrap_as_single_agent(terminateds), self._wrap_as_single_agent(
                truncateds), self._wrap_as_single_agent(step_infos)
        else:
            return obses, rewards, terminateds, truncateds, step_infos

    def close(self):
        if self.engine is not None:
            close_engine()
        if self.logger is not None:
            self.logger.handlers.clear()
            del self.logger
        self.logger = None

    def force_close(self):
        print("Closing environment ... Please wait")
        self.close()
        time.sleep(2)  # Sleep two seconds
        raise KeyboardInterrupt("'Esc' is pressed. MetaDrive exits now.")

    def capture(self, file_name=None):
        if not hasattr(self, "_capture_img"):
            self._capture_img = PNMImage()
        self.engine.win.getScreenshot(self._capture_img)
        if file_name is None:
            file_name = "main_index_{}_step_{}_{}.png".format(self.current_seed, self.engine.episode_step, time.time())
        self._capture_img.write(file_name)
        self.logger.info("Image is saved at: {}".format(file_name))

    def for_each_vehicle(self, func, *args, **kwargs):
        return self.agent_manager.for_each_active_agents(func, *args, **kwargs)

    @property
    def vehicle(self):
        """A helper to return the vehicle only in the single-agent environment!"""
        assert len(self.vehicles) == 1, (
            "env.vehicle is only supported in single-agent environment!"
            if len(self.vehicles) > 1 else "Please initialize the environment first!"
        )
        ego_v = self.vehicles[DEFAULT_AGENT]
        return ego_v

    def get_single_observation(self):
        if self.config["image_observation"]:
            o = ImageStateObservation(self.config)
        else:
            o = LidarStateObservation(self.config)
        return o

    def _wrap_as_single_agent(self, data):
        return data[next(iter(self.vehicles.keys()))]

    def seed(self, seed=None):
        if seed is not None:
            set_global_random_seed(seed)

    @property
    def current_seed(self):
        return self.engine.global_random_seed

    @property
    def observations(self):
        """
        Return observations of active and controllable vehicles
        :return: Dict
        """
        return self.agent_manager.get_observations()

    @property
    def observation_space(self) -> gym.Space:
        """
        Return observation spaces of active and controllable vehicles
        :return: Dict
        """
        ret = self.agent_manager.get_observation_spaces()
        if (not self.is_multi_agent) or self.config["return_single_space"]:
            return next(iter(ret.values()))
        else:
            return gym.spaces.Dict(ret)

    @property
    def action_space(self) -> gym.Space:
        """
        Return observation spaces of active and controllable vehicles
        :return: Dict
        """
        ret = self.agent_manager.get_action_spaces()
        if (not self.is_multi_agent) or self.config["return_single_space"]:
            return next(iter(ret.values()))
        else:
            return gym.spaces.Dict(ret)

    @property
    def vehicles(self):
        """
        Return all active vehicles
        :return: Dict[agent_id:vehicle]
        """
        return self.agent_manager.active_agents

    @property
    def vehicles_including_just_terminated(self):
        """
        Return all vehicles that occupy some space in current environments
        :return: Dict[agent_id:vehicle]
        """
        ret = self.agent_manager.active_agents
        ret.update(self.agent_manager.just_terminated_agents)
        return ret

    def setup_engine(self):
        """
        Engine setting after launching
        """
        self.engine.accept("r", self.reset)
        self.engine.accept("c", self.capture)
        self.engine.register_manager("agent_manager", self.agent_manager)
        self.engine.register_manager("record_manager", RecordManager())
        self.engine.register_manager("replay_manager", ReplayManager())

    @property
    def current_map(self):
        return self.engine.current_map

    def _reset_global_seed(self, force_seed=None):
        current_seed = force_seed if force_seed is not None else get_np_random(None).randint(0, int(1e4))
        self.seed(current_seed)

    @property
    def maps(self):
        return self.engine.map_manager.maps

    def _render_topdown(self, text, *args, **kwargs):
        return self.engine.render_topdown(text, *args, **kwargs)

    @property
    def main_camera(self):
        return self.engine.main_camera

    @property
    def current_track_vehicle(self):
        return self.engine.current_track_vehicle

    @property
    def _top_down_renderer(self):
        return self.engine._top_down_renderer

    @property
    def episode_step(self):
        return self.engine.episode_step if self.engine is not None else 0

    def export_scenarios(
        self,
        policies: Union[dict, Callable],
        scenario_index: Union[list, int],
        max_episode_length=None,
        verbose=False,
        suppress_warning=False,
        render_topdown=False,
        return_done_info=True,
        to_dict=True
    ):
        """
        We export scenarios into a unified format with 10hz sample rate
        """
        def _act(observation):
            if isinstance(policies, dict):
                ret = {}
                for id, o in observation.items():
                    ret[id] = policies[id](o)
            else:
                ret = policies(observation)
            return ret

        if self.is_multi_agent:
            assert isinstance(policies, dict), "In MARL setting, policies should be mapped to agents according to id"
        else:
            assert isinstance(policies, Callable), "In single agent case, policy should be a callable object, taking" \
                                                   "observation as input."
        scenarios_to_export = dict()
        if isinstance(scenario_index, int):
            scenario_index = [scenario_index]
        self.config["record_episode"] = True
        done_info = {}
        for index in scenario_index:
            obs = self.reset(seed=index)
            done = False
            count = 0
            info = None
            while not done:
                obs, reward, terminated, truncated, info = self.step(_act(obs))
                done = terminated or truncated
                count += 1
                if max_episode_length is not None and count > max_episode_length:
                    done = True
                    info[TerminationState.MAX_STEP] = True
                if count > 10000 and not suppress_warning:
                    self.logger.warning(
                        "Episode length is too long! If this behavior is intended, "
                        "set suppress_warning=True to disable this message"
                    )
                if render_topdown:
                    self.render("topdown")
            episode = self.engine.dump_episode()
            if verbose:
                self.logger.info("Finish scenario {} with {} steps.".format(index, count))
            scenarios_to_export[index] = convert_recorded_scenario_exported(episode, to_dict=to_dict)
            done_info[index] = info
        self.config["record_episode"] = False
        if return_done_info:
            return scenarios_to_export, done_info
        else:
            return scenarios_to_export

    def export_single_scenario(self):
        """
        Similar export_scenarios, this function transform the internal recorded frames to a standard
        scenario description.
        """
        episode = self.engine.dump_episode()
        return convert_recorded_scenario_exported(episode)

    @property
    def logger_name(self):
        return self.__class__.__name__



"""
This script aims to convert nuplan scenarios to ScenarioDescription, so that we can load any nuplan scenarios into
MetaDrive.
"""
from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario import CameraChannel, LidarChannel

from metadrive.utils.nuplan.utils import get_nuplan_scenarios

# import os
#
# os.environ["NUPLAN_DATA_STORE"] = "s3"

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)

if __name__ == '__main__':
    raise ValueError("It seems the sensor data is not ready yet")
    from nuplan.common.actor_state.vehicle_parameters import get_pacifica_parameters
    from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario import NuPlanScenario, CameraChannel, LidarChannel
    from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario_utils import ScenarioExtractionInfo

    NUPLAN_DATA_ROOT = "/mnt/data/nuplan/dataset/"
    NUPLAN_MAP_VERSION = "nuplan-maps-v1.0"
    NUPLAN_MAPS_ROOT = "/mnt/data/nuplan/dataset/maps"
    NUPLAN_SENSOR_ROOT = f"/mnt/data/nuplan/dataset/nuplan-v1.1/sensor_blobs"
    TEST_DB_FILE = f"{NUPLAN_DATA_ROOT}/nuplan-v1.1/splits/mini/2021.05.12.22.00.38_veh-35_01008_01518.db"
    MAP_NAME = "us-nv-las-vegas"
    TEST_INITIAL_LIDAR_PC = "58ccd3df9eab54a3"
    TEST_INITIAL_TIMESTAMP = 1620858198150622

    scenario = NuPlanScenario(
        data_root=f"{NUPLAN_DATA_ROOT}/nuplan-v1.1/splits/mini",
        log_file_load_path=TEST_DB_FILE,
        initial_lidar_token=TEST_INITIAL_LIDAR_PC,
        initial_lidar_timestamp=TEST_INITIAL_TIMESTAMP,
        scenario_type="scenario_type",
        map_root=NUPLAN_MAPS_ROOT,
        map_version=NUPLAN_MAP_VERSION,
        map_name=MAP_NAME,
        scenario_extraction_info=ScenarioExtractionInfo(
            scenario_name="scenario_name", scenario_duration=20, extraction_offset=1, subsample_ratio=0.5
        ),
        ego_vehicle_parameters=get_pacifica_parameters(),
        sensor_root=NUPLAN_DATA_ROOT + "/nuplan-v1.1/sensor_blobs",
    )



import logging

try:
    import geopandas as gpd
except ImportError:
    pass

import numpy as np
from shapely.geometry.linestring import LineString
from shapely.geometry.multilinestring import MultiLineString
from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector
from metadrive.component.lane.point_lane import PointLane

from metadrive.utils.interpolating_line import InterpolatingLine
logger = logging.getLogger(__name__)


class NuPlanLane(PointLane):
    def __init__(self, lane_meta_data, nuplan_center, need_lane_localization=False):
        """
        Extract the lane information of one waymo lane, and do coordinate shift
        """

        if isinstance(lane_meta_data.polygon.boundary, MultiLineString):
            # logger.warning("Stop using boundaries! Use exterior instead!")
            boundary = gpd.GeoSeries(lane_meta_data.polygon.boundary).explode(index_parts=True)
            sizes = []
            for idx, polygon in enumerate(boundary[0]):
                sizes.append(len(polygon.xy[1]))
            points = boundary[0][np.argmax(sizes)].xy
        elif isinstance(lane_meta_data.polygon.boundary, LineString):
            points = lane_meta_data.polygon.boundary.xy
        polygon = [[points[0][i], points[1][i]] for i in range(len(points[0]))]
        # polygon += [[points[0][i], points[1][i], 0.] for i in range(len(points[0]))]
        polygon = nuplan_to_metadrive_vector(polygon, nuplan_center=[nuplan_center[0], nuplan_center[1]])
        super(NuPlanLane, self).__init__(
            self._extract_centerline(lane_meta_data, nuplan_center),
            width=None,  # we use width_at to get width
            polygon=polygon,
            need_lane_localization=need_lane_localization
        )
        self.index = lane_meta_data.id
        self.entry_lanes = lane_meta_data.incoming_edges,
        self.exit_lanes = lane_meta_data.outgoing_edges,
        self.left_lanes = lane_meta_data.adjacent_edges[0],
        self.right_lanes = lane_meta_data.adjacent_edges[-1]

        self.left_boundary = InterpolatingLine(self._get_boundary_points(lane_meta_data.left_boundary, nuplan_center))
        self.right_boundary = InterpolatingLine(self._get_boundary_points(lane_meta_data.right_boundary, nuplan_center))
        self.width = self.VIS_LANE_WIDTH

    @staticmethod
    def _extract_centerline(map_obj, nuplan_center):
        path = map_obj.baseline_path.discrete_path
        points = np.array([nuplan_to_metadrive_vector([pose.x, pose.y], nuplan_center) for pose in path])
        return points

    def width_at(self, longitudinal: float) -> float:
        l_pos = self.left_boundary.position(longitudinal, 0)
        r_pos = self.right_boundary.position(longitudinal, 0)
        return min(np.linalg.norm(r_pos - l_pos), self.VIS_LANE_WIDTH)

    def __del__(self):
        logging.debug("NuPlanLane is released")

    def destroy(self):
        self.index = None
        super(NuPlanLane, self).destroy()

    @staticmethod
    def _get_boundary_points(boundary, center):
        path = boundary.discrete_path
        points = np.array([nuplan_to_metadrive_vector([pose.x, pose.y], nuplan_center=center) for pose in path])
        return points


if __name__ == "__main__":
    raise ValueError("Can not be run")
    file_path = AssetLoader.file_path("waymo", "test.pkl", return_raw_style=False)
    data = read_waymo_data(file_path)
    print(data)
    lane = NuPlanLane(108, data["map"])



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# Disable the "exactly one space required after comma" message, for the input
# bindings it looks nicer to insert some spaces (see the setup method)
# pylint: disable=bad-whitespace

from __future__ import print_function

import sys
import os
import subprocess

from panda3d.core import ModifierButtons, Vec3, PStatClient
from panda3d.core import Point3, CurveFitter


class MovementController(object):
    """ This is a helper class, used to controll the camera and enable various
    debugging features. It is not really part of the pipeline, but included to
    view the demo scenes. """
    def __init__(self, showbase):
        self.showbase = showbase
        self.movement = [0, 0, 0]
        self.velocity = Vec3(0.0)
        self.hpr_movement = [0, 0]
        self.speed = 0.4
        self.initial_position = Vec3(0)
        self.initial_destination = Vec3(0)
        self.initial_hpr = Vec3(0)
        self.mouse_enabled = False
        self.last_mouse_pos = [0, 0]
        self.mouse_sensivity = 0.7
        self.keyboard_hpr_speed = 0.4
        self.use_hpr = False
        self.smoothness = 6.0
        self.bobbing_amount = 1.5
        self.bobbing_speed = 0.5

    def set_initial_position(self, pos, target):
        """ Sets the initial camera position """
        self.initial_position = pos
        self.initial_destination = target
        self.use_hpr = False
        self.reset_to_initial()

    def set_initial_position_hpr(self, pos, hpr):
        """ Sets the initial camera position """
        self.initial_position = pos
        self.initial_hpr = hpr
        self.use_hpr = True
        self.reset_to_initial()

    def reset_to_initial(self):
        """ Resets the camera to the initial position """
        self.showbase.camera.set_pos(self.initial_position)

        if self.use_hpr:
            self.showbase.camera.set_hpr(self.initial_hpr)
        else:
            self.showbase.camera.look_at(
                self.initial_destination.x, self.initial_destination.y, self.initial_destination.z
            )

    def set_movement(self, direction, amount):
        self.movement[direction] = amount

    def set_hpr_movement(self, direction, amount):
        self.hpr_movement[direction] = amount

    def set_mouse_enabled(self, enabled):
        self.mouse_enabled = enabled

    def increase_speed(self):
        self.speed *= 1.4

    def decrease_speed(self):
        self.speed *= 0.6

    def unbind(self):
        """ Unbinds the movement controler and restores the previous state """
        raise NotImplementedError()

    @property
    def clock_obj(self):
        """ Returns the global clock object """
        return self.showbase.taskMgr.globalClock

    def setup(self):
        """ Attaches the movement controller and inits the keybindings """
        # x
        self.showbase.accept("raw-w", self.set_movement, [0, 1])
        self.showbase.accept("raw-w-up", self.set_movement, [0, 0])
        self.showbase.accept("raw-s", self.set_movement, [0, -1])
        self.showbase.accept("raw-s-up", self.set_movement, [0, 0])

        # y
        self.showbase.accept("raw-a", self.set_movement, [1, -1])
        self.showbase.accept("raw-a-up", self.set_movement, [1, 0])
        self.showbase.accept("raw-d", self.set_movement, [1, 1])
        self.showbase.accept("raw-d-up", self.set_movement, [1, 0])

        # z
        self.showbase.accept("space", self.set_movement, [2, 1])
        self.showbase.accept("space-up", self.set_movement, [2, 0])
        self.showbase.accept("shift", self.set_movement, [2, -1])
        self.showbase.accept("shift-up", self.set_movement, [2, 0])

        # wireframe + debug + buffer viewer
        self.showbase.accept("f3", self.showbase.toggle_wireframe)
        self.showbase.accept("f11", lambda: self.showbase.win.save_screenshot("screenshot.png"))
        self.showbase.accept("j", self.print_position)

        # mouse
        self.showbase.accept("mouse1", self.set_mouse_enabled, [True])
        self.showbase.accept("mouse1-up", self.set_mouse_enabled, [False])

        # arrow mouse navigation
        self.showbase.accept("arrow_up", self.set_hpr_movement, [1, 1])
        self.showbase.accept("arrow_up-up", self.set_hpr_movement, [1, 0])
        self.showbase.accept("arrow_down", self.set_hpr_movement, [1, -1])
        self.showbase.accept("arrow_down-up", self.set_hpr_movement, [1, 0])
        self.showbase.accept("arrow_left", self.set_hpr_movement, [0, 1])
        self.showbase.accept("arrow_left-up", self.set_hpr_movement, [0, 0])
        self.showbase.accept("arrow_right", self.set_hpr_movement, [0, -1])
        self.showbase.accept("arrow_right-up", self.set_hpr_movement, [0, 0])

        # increase / decrease speed
        self.showbase.accept("+", self.increase_speed)
        self.showbase.accept("-", self.decrease_speed)

        # disable modifier buttons to be able to move while pressing shift for
        # example
        self.showbase.mouseWatcherNode.set_modifier_buttons(ModifierButtons())
        self.showbase.buttonThrowers[0].node().set_modifier_buttons(ModifierButtons())

        # disable pandas builtin mouse control
        self.showbase.disableMouse()

        # add ourself as an update task which gets executed very early before
        # the rendering
        self.update_task = self.showbase.addTask(self.update, "RP_UpdateMovementController", sort=-40)

        # Hotkeys to connect to pstats and reset the initial position
        self.showbase.accept("1", PStatClient.connect)
        self.showbase.accept("3", self.reset_to_initial)

    def print_position(self):
        """ Prints the camera position and hpr """
        pos = self.showbase.cam.get_pos(self.showbase.render)
        hpr = self.showbase.cam.get_hpr(self.showbase.render)
        print("(Vec3({}, {}, {}), Vec3({}, {}, {})),".format(pos.x, pos.y, pos.z, hpr.x, hpr.y, hpr.z))

    def update(self, task):
        """ Internal update method """

        delta = self.clock_obj.get_dt()

        # Update mouse first
        if self.showbase.mouseWatcherNode.has_mouse():
            x = self.showbase.mouseWatcherNode.get_mouse_x()
            y = self.showbase.mouseWatcherNode.get_mouse_y()
            self.current_mouse_pos = (
                x * self.showbase.camLens.get_fov().x * self.mouse_sensivity,
                y * self.showbase.camLens.get_fov().y * self.mouse_sensivity
            )

            if self.mouse_enabled:
                diffx = self.last_mouse_pos[0] - self.current_mouse_pos[0]
                diffy = self.last_mouse_pos[1] - self.current_mouse_pos[1]

                # Don't move in the very beginning
                if self.last_mouse_pos[0] == 0 and self.last_mouse_pos[1] == 0:
                    diffx = 0
                    diffy = 0

                self.showbase.camera.set_h(self.showbase.camera.get_h() + diffx)
                self.showbase.camera.set_p(self.showbase.camera.get_p() - diffy)

            self.last_mouse_pos = self.current_mouse_pos[:]

        # Compute movement in render space
        movement_direction = (Vec3(self.movement[1], self.movement[0], 0) * self.speed * delta * 100.0)

        # transform by the camera direction
        camera_quaternion = self.showbase.camera.get_quat(self.showbase.render)
        translated_direction = camera_quaternion.xform(movement_direction)

        # z-force is inddpendent of camera direction
        translated_direction.add_z(self.movement[2] * delta * 120.0 * self.speed)

        self.velocity += translated_direction * 0.15

        # apply the new position
        self.showbase.camera.set_pos(self.showbase.camera.get_pos() + self.velocity)

        # transform rotation (keyboard keys)
        rotation_speed = self.keyboard_hpr_speed * 100.0
        rotation_speed *= delta
        self.showbase.camera.set_hpr(
            self.showbase.camera.get_hpr() + Vec3(self.hpr_movement[0], self.hpr_movement[1], 0) * rotation_speed
        )

        # fade out velocity
        self.velocity = self.velocity * max(0.0, 1.0 - delta * 60.0 / max(0.01, self.smoothness))

        # bobbing
        ftime = self.clock_obj.get_frame_time()
        rotation = (ftime % self.bobbing_speed) / self.bobbing_speed
        rotation = (min(rotation, 1.0 - rotation) * 2.0 - 0.5) * 2.0
        if self.velocity.length_squared() > 1e-5 and self.speed > 1e-5:
            rotation *= self.bobbing_amount
            rotation *= min(1, self.velocity.length()) / self.speed * 0.5
        else:
            rotation = 0
        self.showbase.camera.set_r(rotation)
        return task.cont

    def play_motion_path(self, points, point_duration=1.2):
        """ Plays a motion path from the given set of points """
        fitter = CurveFitter()
        for i, (pos, hpr) in enumerate(points):
            fitter.add_xyz_hpr(i, pos, hpr)

        fitter.compute_tangents(1.0)
        curve = fitter.make_hermite()
        print("Starting motion path with", len(points), "CVs")

        self.showbase.render2d.hide()
        self.showbase.aspect2d.hide()

        self.curve = curve
        self.curve_time_start = self.clock_obj.get_frame_time()
        self.curve_time_end = self.clock_obj.get_frame_time() + len(points) * point_duration
        self.delta_time_sum = 0.0
        self.delta_time_count = 0
        self.showbase.addTask(self.camera_motion_update, "RP_CameraMotionPath", sort=-50)
        self.showbase.taskMgr.remove(self.update_task)

    def camera_motion_update(self, task):
        if self.clock_obj.get_frame_time() > self.curve_time_end:
            print("Camera motion path finished")

            # Print performance stats
            avg_ms = self.delta_time_sum / self.delta_time_count
            print("Average frame time (ms): {:4.1f}".format(avg_ms * 1000.0))
            print("Average frame rate: {:4.1f}".format(1.0 / avg_ms))

            self.update_task = self.showbase.addTask(self.update, "RP_UpdateMovementController", sort=-50)
            self.showbase.render2d.show()
            self.showbase.aspect2d.show()
            return task.done

        lerp = (self.clock_obj.get_frame_time() - self.curve_time_start) /\
            (self.curve_time_end - self.curve_time_start)
        lerp *= self.curve.get_max_t()

        pos, hpr = Point3(0), Vec3(0)
        self.curve.evaluate_xyz(lerp, pos)
        self.curve.evaluate_hpr(lerp, hpr)

        self.showbase.camera.set_pos(pos)
        self.showbase.camera.set_hpr(hpr)

        self.delta_time_sum += self.clock_obj.get_dt()
        self.delta_time_count += 1

        return task.cont



import math
from dataclasses import dataclass

import numpy as np

from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector
from metadrive.utils.coordinates_shift import panda_heading

try:
    import geopandas as gpd
    from nuplan.common.actor_state.state_representation import Point2D
    from nuplan.common.maps.maps_datatypes import SemanticMapLayer, StopLineType
    from shapely.ops import unary_union
except ImportError:
    pass

from metadrive.component.block.base_block import BaseBlock
from metadrive.component.lane.nuplan_lane import NuPlanLane
from metadrive.component.road_network.edge_road_network import EdgeRoadNetwork
from metadrive.constants import DrivableAreaProperty
from metadrive.constants import PGLineColor, PGLineType
from metadrive.utils.interpolating_line import InterpolatingLine
from metadrive.utils.math import wrap_to_pi, norm


@dataclass
class LaneLineProperty:
    id: str
    points: list
    color: PGLineColor
    type: PGLineType
    in_road_connector: bool


class NuPlanBlock(BaseBlock):
    def __init__(self, block_index: int, global_network, random_seed, map_index, nuplan_center):
        self.map_index = map_index
        self.nuplan_center = nuplan_center
        super(NuPlanBlock, self).__init__(block_index, global_network, random_seed)

        # authorize engine access for this object
        # self.engine = get_engine()
        self._nuplan_map_api = self.engine.data_manager.current_scenario.map_api

        self.lines = {}
        # self.boundaries = {}

    @property
    def map_api(self):
        return self._nuplan_map_api

    def _sample_topology(self) -> bool:
        """
        This function is modified from _render_map in nuplan-devkit.simulation_tile.py, it should be setting outside now
        """

        return True
        # """
        # The following content is deprecated
        # """
        # # raise ValueError("Deprecated")
        # # Deprecated
        # map_api = self._nuplan_map_api
        # # Center is Important !
        # center = self.nuplan_center
        # layer_names = [
        #     SemanticMapLayer.LANE_CONNECTOR,
        #     SemanticMapLayer.LANE,
        #     SemanticMapLayer.CROSSWALK,
        #     SemanticMapLayer.INTERSECTION,
        #     SemanticMapLayer.STOP_LINE,
        #     SemanticMapLayer.WALKWAYS,
        #     SemanticMapLayer.CARPARK_AREA,
        #     SemanticMapLayer.ROADBLOCK,
        #     SemanticMapLayer.ROADBLOCK_CONNECTOR,
        #
        #     # unsupported yet
        #     # SemanticMapLayer.STOP_SIGN,
        #     # SemanticMapLayer.DRIVABLE_AREA,
        # ]
        #
        # nearest_vector_map = map_api.get_proximal_map_objects(Point2D(*center), self._radius, layer_names)
        # # Filter out stop polygons in turn stop
        # if SemanticMapLayer.STOP_LINE in nearest_vector_map:
        #     stop_polygons = nearest_vector_map[SemanticMapLayer.STOP_LINE]
        #     nearest_vector_map[SemanticMapLayer.STOP_LINE] = [
        #         stop_polygon for stop_polygon in stop_polygons if stop_polygon.stop_line_type != StopLineType.TURN_STOP
        #     ]
        #
        # block_polygons = []
        # # Lane and lane line
        # for layer in [SemanticMapLayer.ROADBLOCK, SemanticMapLayer.ROADBLOCK_CONNECTOR]:
        #     for block in nearest_vector_map[layer]:
        #         for lane in block.interior_edges:
        #             if hasattr(lane, "baseline_path"):
        #                 self.block_network.add_lane(NuPlanLane(nuplan_center=center, lane_meta_data=lane))
        #                 is_connector = True if layer == SemanticMapLayer.ROADBLOCK_CONNECTOR else False
        #                 self._get_lane_line(lane, is_road_connector=is_connector, nuplan_center=center)
        #         if layer == SemanticMapLayer.ROADBLOCK:
        #             block_polygons.append(block.polygon)
        #
        # # intersection road connector
        # interpolygons = [block.polygon for block in nearest_vector_map[SemanticMapLayer.INTERSECTION]]
        # boundaries = gpd.GeoSeries(unary_union(interpolygons + block_polygons)).boundary.explode(index_parts=True)
        # # boundaries.plot()
        # # plt.show()
        # for idx, boundary in enumerate(boundaries[0]):
        #     block_points = np.array(list(i for i in zip(boundary.coords.xy[0], boundary.coords.xy[1])))
        #     block_points -= center
        #     self.lines["boundary_{}".format(idx)] = LaneLineProperty(
        #         block_points, PGLineColor.GREY, PGLineType.CONTINUOUS, in_road_connector=False
        #     )

    def create_in_world(self):
        """
        The lane line should be created separately
        """
        graph = self.block_network.graph
        for id, lane_info in graph.items():
            lane = lane_info.lane
            lane.construct_lane_in_block(self, lane_index=id)
            # lane.construct_lane_line_in_block(self, [True if len(lane.left_lanes) == 0 else False,
            #                                          True if len(lane.right_lanes) == 0 else False, ])
        # for line_prop in self.lines.values():
        #     self.construct_nuplan_continuous_line(line_prop.points, line_prop.color)

        for line_prop in self.lines.values():
            if line_prop.type == PGLineType.CONTINUOUS:
                self.construct_nuplan_continuous_line(line_prop.points, line_prop.color)
            if line_prop.type == PGLineType.BROKEN:
                self.construct_nuplan_broken_line(line_prop.points, line_prop.color)

    def construct_nuplan_continuous_line(self, polyline, color):
        for start, end, in zip(polyline[:-1], polyline[1:]):
            node_path_list = NuPlanLane.construct_lane_line_segment(self, start, end, color, PGLineType.CONTINUOUS)
            self._node_path_list.extend(node_path_list)

    def construct_nuplan_broken_line(self, polyline, color):
        line = InterpolatingLine(polyline)
        segment_num = int(line.length / (2 * DrivableAreaProperty.STRIPE_LENGTH))
        for segment in range(segment_num):
            start = line.get_point(segment * DrivableAreaProperty.STRIPE_LENGTH * 2)
            end = line.get_point(segment * DrivableAreaProperty.STRIPE_LENGTH * 2 + DrivableAreaProperty.STRIPE_LENGTH)
            if segment == segment_num - 1:
                end = line.get_point(line.length - DrivableAreaProperty.STRIPE_LENGTH)
            node_path_list = NuPlanLane.construct_lane_line_segment(self, start, end, color, PGLineType.BROKEN)
            self._node_path_list.extend(node_path_list)

    def construct_nuplan_sidewalk(self, polyline):
        line = InterpolatingLine(polyline)
        seg_len = DrivableAreaProperty.LANE_SEGMENT_LENGTH
        segment_num = int(line.length / seg_len)
        last_theta = None
        for segment in range(segment_num):
            lane_start = line.get_point(segment * seg_len)
            lane_end = line.get_point((segment + 1) * seg_len)
            if segment == segment_num - 1:
                lane_end = line.get_point(line.length)
            direction_v = lane_end - lane_start
            theta = panda_heading(math.atan2(direction_v[1], direction_v[0]))
            if segment == segment_num - 1:
                factor = 1
            else:
                factor = 1.25
                if last_theta is not None:
                    diff = wrap_to_pi(theta) - wrap_to_pi(last_theta)
                    if diff > 0:
                        factor += math.sin(abs(diff) / 2) * DrivableAreaProperty.SIDEWALK_WIDTH / norm(
                            lane_start[0] - lane_end[0], lane_start[1] - lane_end[1]
                        ) + 0.15
                    else:
                        factor -= math.sin(abs(diff) / 2) * DrivableAreaProperty.SIDEWALK_WIDTH / norm(
                            lane_start[0] - lane_end[0], lane_start[1] - lane_end[1]
                        )
            last_theta = theta
            node_path_list = NuPlanLane.construct_sidewalk_segment(
                self, lane_start, lane_end, length_multiply=factor, extra_thrust=1
            )
            self._node_path_list.extend(node_path_list)

    @property
    def block_network_type(self):
        return EdgeRoadNetwork

    def destroy(self):
        self.map_index = None
        self.nuplan_map_data = None
        super(NuPlanBlock, self).destroy()

    def __del__(self):
        self.destroy()
        super(NuPlanBlock, self).__del__()
        # print("NuPlan Block is being deleted.")

    @staticmethod
    def _get_points_from_boundary(boundary, center):
        path = boundary.discrete_path
        points = [(pose.x, pose.y) for pose in path]
        points = nuplan_to_metadrive_vector(points, center)
        return points

    def set_lane_line(self, lane, nuplan_center, is_road_connector=False):
        center = nuplan_center
        boundaries = self.map_api._get_vector_map_layer(SemanticMapLayer.BOUNDARIES)

        if not is_road_connector:

            left = lane.left_boundary
            if left.id not in self.lines:
                line_type = self._metadrive_line_type(int(boundaries.loc[[str(left.id)]]["boundary_type_fid"]))
                line_color = PGLineColor.YELLOW if lane.adjacent_edges[0] is None and lane.adjacent_edges[
                    1] is not None else PGLineColor.GREY
                if line_color == PGLineColor.YELLOW:
                    self.lines[left.id] = LaneLineProperty(
                        left.id,
                        self._get_points_from_boundary(left, center),
                        line_color,
                        line_type,
                        in_road_connector=is_road_connector
                    )
                elif line_type == PGLineType.BROKEN:
                    self.lines[left.id] = LaneLineProperty(
                        left.id,
                        self._get_points_from_boundary(left, center),
                        PGLineColor.GREY,
                        line_type,
                        in_road_connector=is_road_connector
                    )

            # right = lane.right_boundary
            # if right.id not in self.lines:
            #     line_type = self._metadrive_line_type(int(boundaries.loc[[str(right.id)]]["boundary_type_fid"]))
            #     if line_type == PGLineType.BROKEN:
            #         self.lines[right.id] = LaneLineProperty(
            #             self._get_points_from_boundary(right, center), PGLineColor.GREY, line_type,
            #             in_road_connector=is_road_connector
            #         )

    def _get_sidewalk(self, walkway, nuplan_center):
        center = nuplan_center
        if walkway.id not in self.lines:
            points = np.array(
                [
                    i for i in zip(
                        walkway.polygon.boundary.coords.xy[0] - center[0], walkway.polygon.boundary.coords.xy[1] -
                        center[1]
                    )
                ]
            )
            self.lines[walkway.id] = LaneLineProperty(
                walkway.id, points, PGLineColor.GREY, PGLineType.SIDE, in_road_connector=False
            )

    def _get_lane_boundary(self, line):
        pass

    def _metadrive_line_type(self, nuplan_type):
        if nuplan_type == 2:
            return PGLineType.CONTINUOUS
        elif nuplan_type == 0:
            return PGLineType.BROKEN
        elif nuplan_type == 3:
            return PGLineType.NONE
        else:
            raise ValueError("Unknown line tyep: {}".format(nuplan_type))



from typing import Union

from panda3d.core import NodePath

from metadrive.component.algorithm.BIG import BIG
from metadrive.component.algorithm.blocks_prob_dist import PGBlockDistConfig
from metadrive.component.map.base_map import BaseMap
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.pg_block import PGBlock
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.engine.core.physics_world import PhysicsWorld


class NextStep:
    back = 0
    forward = 1
    search_sibling = 3
    destruct_current = 4


class BigGenerateMethod:
    BLOCK_SEQUENCE = "block_sequence"
    BLOCK_NUM = "block_num"


class CityBIG(BIG):
    MAX_TRIAL = 2

    def __init__(
        self,
        lane_num: int,
        lane_width: float,
        global_network: NodeRoadNetwork,
        render_node_path: NodePath,
        physics_world: PhysicsWorld,
        # block_type_version: str,
        random_seed=None
    ):
        super(CityBIG, self).__init__(
            lane_num,
            lane_width,
            global_network,
            render_node_path,
            physics_world,
            # block_type_version,
            random_seed=random_seed
        )

    def generate(self, generate_method: BigGenerateMethod, parameter: Union[str, int]):
        """
        In order to embed it to the show_base loop, we implement BIG in a more complex way
        """
        if generate_method == BigGenerateMethod.BLOCK_NUM:
            assert isinstance(parameter, int), "When generating map by assigning block num, the parameter should be int"
            self.block_num = parameter + 1
        elif generate_method == BigGenerateMethod.BLOCK_SEQUENCE:
            assert isinstance(parameter, str), "When generating map from block sequence, the parameter should be a str"
            self.block_num = len(parameter) + 1
            self._block_sequence = FirstPGBlock.ID + parameter
        while True:
            if self.big_helper_func():
                break
        return self._global_network

    def sample_block(self) -> PGBlock:
        """
        Sample a random block type
        """
        if self._block_sequence is None:
            block_types = PGBlockDistConfig.all_blocks()
            block_probabilities = PGBlockDistConfig.block_probability()
            block_type = self.np_random.choice(block_types, p=block_probabilities)
            from metadrive.utils.registry import get_metadrive_class
            block_type = get_metadrive_class(block_type)
        else:
            type_id = self._block_sequence[len(self.blocks)]
            block_type = PGBlockDistConfig.get_block(type_id)

        # exclude first block
        socket_used = set([block.pre_block_socket for block in self.blocks[1:]])
        socket_available = []
        for b in self.blocks:
            socket_available += b.get_socket_list()
        socket_available = set(socket_available).difference(socket_used)
        socket = self.np_random.choice(sorted(list(socket_available), key=lambda x: x.index))

        block = block_type(
            len(self.blocks),
            socket,
            self._global_network,
            self.np_random.randint(0, 10000),
            ignore_intersection_checking=False
        )
        return block


class CityMap(BaseMap):
    def _generate(self):
        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        big_map = CityBIG(
            self._config[self.LANE_NUM],
            self._config[self.LANE_WIDTH],
            self.road_network,
            parent_node_path,
            physics_world,
            # self._config["block_type_version"]
        )
        big_map.generate(self._config[self.GENERATE_TYPE], self._config[self.GENERATE_CONFIG])
        self.blocks = big_map.blocks

    @property
    def road_network_type(self):
        return NodeRoadNetwork



import copy
from typing import Union

import numpy as np

from metadrive.component.algorithm.blocks_prob_dist import PGBlockDistConfig
from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import parse_map_config, MapGenerateMethod
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import DEFAULT_AGENT, TerminationState
from metadrive.envs.base_env import BaseEnv
from metadrive.manager.traffic_manager import TrafficMode
from metadrive.utils import clip, Config, get_np_random

METADRIVE_DEFAULT_CONFIG = dict(
    # ===== Generalization =====
    start_seed=0,
    num_scenarios=1,
    environment_num=-1,  # This key is deprecated, use num_scenarios instead!

    # ===== Map Config =====
    map=3,  # int or string: an easy way to fill map_config
    block_dist_config=PGBlockDistConfig,
    random_lane_width=False,
    random_lane_num=False,
    map_config={
        BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
        BaseMap.GENERATE_CONFIG: None,  # it can be a file path / block num / block ID sequence
        BaseMap.LANE_WIDTH: 3.5,
        BaseMap.LANE_NUM: 3,
        "exit_length": 50,
    },
    store_map=True,

    # ===== Traffic =====
    traffic_density=0.1,
    need_inverse_traffic=False,
    traffic_mode=TrafficMode.Trigger,  # "Respawn", "Trigger"
    random_traffic=False,  # Traffic is randomized at default.
    # this will update the vehicle_config and set to traffic
    traffic_vehicle_config=dict(
        show_navi_mark=False,
        show_dest_mark=False,
        enable_reverse=False,
        show_lidar=False,
        show_lane_line_detector=False,
        show_side_detector=False,
    ),

    # ===== Object =====
    accident_prob=0.,  # accident may happen on each block with this probability, except multi-exits block
    static_traffic_object=True,  # object won't react to any collisions

    # ===== Others =====
    use_AI_protector=False,
    save_level=0.5,
    is_multi_agent=False,
    vehicle_config=dict(spawn_lane_index=(FirstPGBlock.NODE_1, FirstPGBlock.NODE_2, 0)),

    # ===== Agent =====
    random_spawn_lane_index=True,
    target_vehicle_configs={
        DEFAULT_AGENT: dict(
            use_special_color=True,
            spawn_lane_index=(FirstPGBlock.NODE_1, FirstPGBlock.NODE_2, 0),
        )
    },

    # ===== Reward Scheme =====
    # See: https://github.com/metadriverse/metadrive/issues/283
    success_reward=10.0,
    out_of_road_penalty=5.0,
    crash_vehicle_penalty=5.0,
    crash_object_penalty=5.0,
    driving_reward=1.0,
    speed_reward=0.1,
    use_lateral_reward=False,

    # ===== Cost Scheme =====
    crash_vehicle_cost=1.0,
    crash_object_cost=1.0,
    out_of_road_cost=1.0,

    # ===== Termination Scheme =====
    out_of_route_done=False,
    on_continuous_line_done=True,
    crash_vehicle_done=True,
    crash_object_done=True,
)


class MetaDriveEnv(BaseEnv):
    @classmethod
    def default_config(cls) -> Config:
        config = super(MetaDriveEnv, cls).default_config()
        config.update(METADRIVE_DEFAULT_CONFIG)
        config.register_type("map", str, int)
        config["map_config"].register_type("config", None)
        return config

    def __init__(self, config: Union[dict, None] = None):
        self.default_config_copy = Config(self.default_config(), unchangeable=True)
        super(MetaDriveEnv, self).__init__(config)

        # map setting
        self.start_seed = self.config["start_seed"]
        self.env_num = self.config["num_scenarios"]
        self.in_stop = False

    def _post_process_config(self, config):
        config = super(MetaDriveEnv, self)._post_process_config(config)
        if not config["rgb_clip"]:
            self.logger.warning(
                "You have set rgb_clip = False, which means the observation will be uint8 values in [0, 255]. "
                "Please make sure you have parsed them later before feeding them to network!"
            )
        if config["environment_num"] != -1:
            self.logger.warning("environment_num is deprecated. Use num_scenarios instead!")
            assert config["num_scenarios"] == 1
            config["num_scenarios"] = config["environment_num"]

        config["map_config"] = parse_map_config(
            easy_map_config=config["map"], new_map_config=config["map_config"], default_config=self.default_config_copy
        )
        config["vehicle_config"]["rgb_clip"] = config["rgb_clip"]
        config["vehicle_config"]["random_agent_model"] = config["random_agent_model"]
        if config.get("gaussian_noise", 0) > 0:
            assert config["vehicle_config"]["lidar"]["gaussian_noise"] == 0, "You already provide config!"
            assert config["vehicle_config"]["side_detector"]["gaussian_noise"] == 0, "You already provide config!"
            assert config["vehicle_config"]["lane_line_detector"]["gaussian_noise"] == 0, "You already provide config!"
            config["vehicle_config"]["lidar"]["gaussian_noise"] = config["gaussian_noise"]
            config["vehicle_config"]["side_detector"]["gaussian_noise"] = config["gaussian_noise"]
            config["vehicle_config"]["lane_line_detector"]["gaussian_noise"] = config["gaussian_noise"]
        if config.get("dropout_prob", 0) > 0:
            assert config["vehicle_config"]["lidar"]["dropout_prob"] == 0, "You already provide config!"
            assert config["vehicle_config"]["side_detector"]["dropout_prob"] == 0, "You already provide config!"
            assert config["vehicle_config"]["lane_line_detector"]["dropout_prob"] == 0, "You already provide config!"
            config["vehicle_config"]["lidar"]["dropout_prob"] = config["dropout_prob"]
            config["vehicle_config"]["side_detector"]["dropout_prob"] = config["dropout_prob"]
            config["vehicle_config"]["lane_line_detector"]["dropout_prob"] = config["dropout_prob"]
        target_v_config = copy.deepcopy(config["vehicle_config"])
        if not config["is_multi_agent"]:
            target_v_config.update(config["target_vehicle_configs"][DEFAULT_AGENT])
            config["target_vehicle_configs"][DEFAULT_AGENT] = target_v_config
        return config

    def _get_observations(self):
        return {DEFAULT_AGENT: self.get_single_observation()}

    def done_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        done = False
        done_info = {
            TerminationState.CRASH_VEHICLE: False,
            TerminationState.CRASH_OBJECT: False,
            TerminationState.CRASH_BUILDING: False,
            TerminationState.OUT_OF_ROAD: False,
            TerminationState.SUCCESS: False,
            TerminationState.MAX_STEP: False,
            # TerminationState.CURRENT_BLOCK: self.vehicle.navigation.current_road.block_ID(),
            TerminationState.ENV_SEED: self.current_seed,
            # crash_vehicle=False, crash_object=False, crash_building=False, out_of_road=False, arrive_dest=False,
        }
        if self._is_arrive_destination(vehicle):
            done = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: arrive_dest.".format(self.current_seed))
            done_info[TerminationState.SUCCESS] = True
        if self._is_out_of_road(vehicle):
            done = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: out_of_road.".format(self.current_seed))
            done_info[TerminationState.OUT_OF_ROAD] = True
        if vehicle.crash_vehicle and self.config["crash_vehicle_done"]:
            done = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: crash vehicle ".format(self.current_seed))
            done_info[TerminationState.CRASH_VEHICLE] = True
        if vehicle.crash_object and self.config["crash_object_done"]:
            done = True
            done_info[TerminationState.CRASH_OBJECT] = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: crash object ".format(self.current_seed))
        if vehicle.crash_building:
            done = True
            done_info[TerminationState.CRASH_BUILDING] = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: crash building ".format(self.current_seed))
        if self.config["max_step_per_agent"] is not None and \
                self.episode_lengths[vehicle_id] >= self.config["max_step_per_agent"]:
            done = True
            done_info[TerminationState.MAX_STEP] = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: max step ".format(self.current_seed))

        if self.config["horizon"] is not None and \
                self.episode_lengths[vehicle_id] >= self.config["horizon"] and not self.is_multi_agent:
            # single agent horizon has the same meaning as max_step_per_agent
            done = True
            done_info[TerminationState.MAX_STEP] = True
            self.logger.info("Episode ended! Scenario Index: {} Reason: max step ".format(self.current_seed))

        # for compatibility
        # crash almost equals to crashing with vehicles
        done_info[TerminationState.CRASH] = (
            done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.CRASH_OBJECT]
            or done_info[TerminationState.CRASH_BUILDING]
        )
        return done, done_info

    def cost_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()
        step_info["cost"] = 0
        if self._is_out_of_road(vehicle):
            step_info["cost"] = self.config["out_of_road_cost"]
        elif vehicle.crash_vehicle:
            step_info["cost"] = self.config["crash_vehicle_cost"]
        elif vehicle.crash_object:
            step_info["cost"] = self.config["crash_object_cost"]
        return step_info['cost'], step_info

    def _is_arrive_destination(self, vehicle):
        long, lat = vehicle.navigation.final_lane.local_coordinates(vehicle.position)
        flag = (vehicle.navigation.final_lane.length - 5 < long < vehicle.navigation.final_lane.length + 5) and (
            vehicle.navigation.get_current_lane_width() / 2 >= lat >=
            (0.5 - vehicle.navigation.get_current_lane_num()) * vehicle.navigation.get_current_lane_width()
        )
        return flag

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        # return vehicle.on_yellow_continuous_line or (not vehicle.on_lane) or vehicle.crash_sidewalk
        ret = not vehicle.on_lane
        if self.config["out_of_route_done"]:
            ret = ret or vehicle.out_of_route
        elif self.config["on_continuous_line_done"]:
            ret = ret or vehicle.on_yellow_continuous_line or vehicle.on_white_continuous_line or vehicle.crash_sidewalk
        return ret

    def reward_function(self, vehicle_id: str):
        """
        Override this func to get a new reward function
        :param vehicle_id: id of BaseVehicle
        :return: reward
        """
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()

        # Reward for moving forward in current lane
        if vehicle.lane in vehicle.navigation.current_ref_lanes:
            current_lane = vehicle.lane
            positive_road = 1
        else:
            current_lane = vehicle.navigation.current_ref_lanes[0]
            current_road = vehicle.navigation.current_road
            positive_road = 1 if not current_road.is_negative_road() else -1
        long_last, _ = current_lane.local_coordinates(vehicle.last_position)
        long_now, lateral_now = current_lane.local_coordinates(vehicle.position)

        # reward for lane keeping, without it vehicle can learn to overtake but fail to keep in lane
        if self.config["use_lateral_reward"]:
            lateral_factor = clip(1 - 2 * abs(lateral_now) / vehicle.navigation.get_current_lane_width(), 0.0, 1.0)
        else:
            lateral_factor = 1.0

        reward = 0.0
        reward += self.config["driving_reward"] * (long_now - long_last) * lateral_factor * positive_road
        reward += self.config["speed_reward"] * (vehicle.speed_km_h / vehicle.max_speed_km_h) * positive_road

        step_info["step_reward"] = reward

        if self._is_arrive_destination(vehicle):
            reward = +self.config["success_reward"]
        elif self._is_out_of_road(vehicle):
            reward = -self.config["out_of_road_penalty"]
        elif vehicle.crash_vehicle:
            reward = -self.config["crash_vehicle_penalty"]
        elif vehicle.crash_object:
            reward = -self.config["crash_object_penalty"]
        return reward, step_info

    def switch_to_third_person_view(self) -> (str, BaseVehicle):
        if self.main_camera is None:
            return
        self.main_camera.reset()
        if self.config["prefer_track_agent"] is not None and self.config["prefer_track_agent"] in self.vehicles.keys():
            new_v = self.vehicles[self.config["prefer_track_agent"]]
            current_track_vehicle = new_v
        else:
            if self.main_camera.is_bird_view_camera():
                current_track_vehicle = self.current_track_vehicle
            else:
                vehicles = list(self.engine.agents.values())
                if self.current_track_vehicle in vehicles:
                    vehicles.remove(self.current_track_vehicle)
                if len(vehicles) < 1:
                    return
                new_v = get_np_random().choice(vehicles)
                current_track_vehicle = new_v
        self.main_camera.track(current_track_vehicle)
        return

    def switch_to_top_down_view(self):
        self.main_camera.stop_track()

    def stop(self):
        self.in_stop = not self.in_stop

    def step(self, *args, **kwargs):
        ret = super(MetaDriveEnv, self).step(*args, **kwargs)
        while self.in_stop:
            self.engine.taskMgr.step()
        return ret

    def next_seed_reset(self):
        if self.current_seed + 1 < self.start_seed + self.env_num:
            self.reset(self.current_seed + 1)
        else:
            self.logger.warning("Can't load next scenario! current seed is already the max scenario index")

    def last_seed_reset(self):
        if self.current_seed - 1 >= self.start_seed:
            self.reset(self.current_seed - 1)
        else:
            self.logger.warning("Can't load last scenario! current seed is already the min scenario index")

    def setup_engine(self):
        super(MetaDriveEnv, self).setup_engine()
        self.engine.accept("b", self.switch_to_top_down_view)
        self.engine.accept("q", self.switch_to_third_person_view)
        self.engine.accept("]", self.next_seed_reset)
        self.engine.accept("[", self.last_seed_reset)
        self.engine.accept("p", self.stop)
        from metadrive.manager.traffic_manager import PGTrafficManager
        from metadrive.manager.pg_map_manager import PGMapManager
        from metadrive.manager.object_manager import TrafficObjectManager
        self.engine.register_manager("map_manager", PGMapManager())
        self.engine.register_manager("traffic_manager", PGTrafficManager())
        if abs(self.config["accident_prob"] - 0) > 1e-2:
            self.engine.register_manager("object_manager", TrafficObjectManager())

    def _reset_global_seed(self, force_seed=None):
        current_seed = force_seed if force_seed is not None else \
            get_np_random(self._DEBUG_RANDOM_SEED).randint(self.start_seed, self.start_seed + self.env_num)
        assert self.start_seed <= current_seed < self.start_seed + self.env_num, \
            "scenario_index (seed) should be in [{}:{})".format(self.start_seed, self.start_seed + self.env_num)
        self.seed(current_seed)


if __name__ == '__main__':

    def _act(env, action):
        assert env.action_space.contains(action)
        obs, reward, terminated, truncated, info = env.step(action)
        assert env.observation_space.contains(obs)
        assert np.isscalar(reward)
        assert isinstance(info, dict)

    env = MetaDriveEnv()
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        _act(env, env.action_space.sample())
        for x in [-1, 0, 1]:
            env.reset()
            for y in [-1, 0, 1]:
                _act(env, [x, y])
    finally:
        env.close()
